#!/usr/bin/env python3

# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020-2021 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

"""Command line tool to start Plom servers."""

__copyright__ = "Copyright (C) 2020-2021 Andrew Rechnitzer, Colin B. Macdonald et al"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import argparse
import locale
import os
import shlex
import shutil
import subprocess
from pathlib import Path
from textwrap import fill, dedent

import pkg_resources

from plom import __version__
from plom import specdir


server_instructions = """Overview of running the Plom server:

  0. Decide on a working directory for the server and cd into it.

  1. Copy the `{specdir}` directory (not just its contents) to your
     server directory.

  2. Run '%(prog)s init' - this will check that everything is in place
     and create necessary sub-directories *and* create config files for
     you to edit.

  3. Run '%(prog)s users' - This will create a template user list
     file for you to edit.  Passwords are displayed in plain text.
     Running with '--demo' option creates a (standard) demo user list,
     while '--auto N' makes an random-generated list of N users.  Edit
     as you see fit.

  4. Run '%(prog)s users <filename>' - This parses the plain-text
     user list, performs some simple sanity checks and then hashes the
     passwords to a new file.

       4a. Optionally you can now delete the file containing
           plain-text passwords.

  5. Now you can start the server with '%(prog)s launch'.

FUTURE - '%(prog)s stop' will stop the server.
""".format(
    specdir=specdir
)


class PlomServerConfigurationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


def checkSpecAndDatabase():
    if os.path.isdir(specdir):
        print("Directory '{}' is present.".format(specdir))
    else:
        print(
            "Cannot find '{}' directory - you must copy this into place before running server. Cannot continue.".format(
                specdir
            )
        )
        exit(1)

    if os.path.isfile(Path(specdir) / "verifiedSpec.toml"):
        print("Test specification present.")
    else:
        print(
            "Cannot find the test specification. Have you run 'plom-build' yet?. Aborting."
        )
        exit(1)

    if os.path.isfile(Path(specdir) / "plom.db"):
        print("Database present: using existing database.")
    else:
        print("Database not yet present: it will be created on first run.")
        # TODO: or should `plom-server init` create it?")

    if os.path.isfile(Path(specdir) / "classlist.csv"):
        print("Classlist present.")
    else:
        print("Cannot find the classlist: expect it later...")


def buildRequiredDirectories():
    lst = [
        "pages",
        "pages/discardedPages",
        "pages/collidingPages",
        "pages/unknownPages",
        "pages/originalPages",
        "markedQuestions",
        "markedQuestions/plomFiles",
        "markedQuestions/commentFiles",
        "serverConfiguration",
    ]
    for dir in lst:
        os.makedirs(dir, exist_ok=True)


def buildSSLKeys():
    """Make new key and cert files if they do not yet exist."""
    key = Path("serverConfiguration") / "plom.key"
    cert = Path("serverConfiguration") / "plom-selfsigned.crt"
    if os.path.isfile(key) and os.path.isfile(cert):
        print("SSL key and certificate already exist - will not change.")
        return

    # Generate new self-signed key/cert
    sslcmd = "openssl req -x509 -sha256 -newkey rsa:2048"
    sslcmd += " -keyout {} -nodes -out {} -days 1000 -subj".format(key, cert)

    # TODO: is this the way to get two digit country code?
    tmp = locale.getdefaultlocale()[0]
    if tmp:
        twodigcc = tmp[-2:]
    else:
        twodigcc = "CA"
    sslcmd += " '/C={}/ST=./L=./CN=localhost'".format(twodigcc)
    try:
        subprocess.check_call(shlex.split(sslcmd))
    except Exception as err:
        raise PlomServerConfigurationError(
            "Something went wrong building ssl keys.\n{}\nCannot continue.".format(err)
        )


def createServerConfig():
    sd = os.path.join("serverConfiguration", "serverDetails.toml")
    if os.path.isfile(sd):
        print("Server config already exists - will not change.")
        return

    template = pkg_resources.resource_string("plom", "serverDetails.toml")
    with open(os.path.join(sd), "wb") as fh:
        fh.write(template)
    print(
        "Please update '{}' with the correct name (or IP) of your server and the port.".format(
            sd
        )
    )


def createBlankPredictions():
    pl = Path(specdir) / "predictionlist.csv"
    if os.path.isfile(pl):
        print("Predictionlist already present.")
        return
    print(
        "Predictionlist will be updated when you run ID-prediction from manager-client."
    )
    with open(pl, "w") as fh:
        fh.write("test, id\n")


def doLatexChecks():
    from plom.textools import texFragmentToPNG
    from plom.server import pageNotSubmitted

    os.makedirs("pleaseCheck", exist_ok=True)

    # check build of fragment
    cdir = os.getcwd()
    keepfiles = ("checkThing.png", "pns.0.0.0.png")
    ct = os.path.join(cdir, "pleaseCheck", keepfiles[0])
    pns = os.path.join(cdir, specdir, "pageNotSubmitted.pdf")
    qns = os.path.join(cdir, specdir, "questionNotSubmitted.pdf")

    fragment = r"\( \mathbb{Z} / \mathbb{Q} \) The cat sat on the mat and verified \LaTeX\ worked okay for plom."

    if not texFragmentToPNG(fragment, ct):
        raise PlomServerConfigurationError(
            "Error latex'ing fragment. Please check your latex distribution."
        )

    # build template pageNotSubmitted.pdf just in case needed
    if not pageNotSubmitted.build_not_submitted_page(pns):
        raise PlomServerConfigurationError(
            "Error building 'pageNotSubmitted.pdf' template page. Please check your latex distribution."
        )
    # build template pageNotSubmitted.pdf just in case needed
    if not pageNotSubmitted.build_not_submitted_question(qns):
        raise PlomServerConfigurationError(
            "Error building 'questionNotSubmitted.pdf' template page. Please check your latex distribution."
        )

    # Try building a replacement for missing page.
    if not pageNotSubmitted.build_test_page_substitute(0, 0, 0):
        raise PlomServerConfigurationError(
            "Error building replacement for missing test page."
        )
    # Try building a replacement for missing page.
    if not pageNotSubmitted.build_homework_question_substitute(0, 0):
        raise PlomServerConfigurationError(
            "Error building replacement for missing homework question."
        )

    shutil.move(keepfiles[1], os.path.join("pleaseCheck", keepfiles[1]))
    print(
        fill(
            dedent(
                """
                Simple latex checks done.  If you feel the need, then please
                examine '{}' and '{}' in the directory 'pleaseCheck'.  The
                first should be a short latex'd fragment with some mathematics
                and text, while the second should be a mostly blank page with
                'page not submitted' stamped across it.  It is safe delete
                both files and the directory.
                """.format(
                    *keepfiles
                )
            )
        )
    )


def initialiseServer():
    print("Do simple existence checks on required files.")
    checkSpecAndDatabase()
    print("Build required directories")
    buildRequiredDirectories()
    print("Building self-signed ssl keys for server")
    buildSSLKeys()
    print("Copy server networking configuration template into place.")
    createServerConfig()
    print("Build blank predictionlist for identifying.")
    createBlankPredictions()
    print(
        "Do latex checks and build 'pageNotSubmitted.pdf', 'questionNotSubmitted.pdf' in case needed"
    )
    doLatexChecks()


#################


def processUsers(userFile, demo, auto):
    # if we have been passed a userFile then process it and return
    if userFile:
        print("Processing user file '{}' to 'userList.json'".format(userFile))
        if os.path.isfile(Path("serverConfiguration") / "userList.json"):
            print("WARNING - this will overwrite the existing userList.json file.")
        from plom.server import manageUserFiles

        manageUserFiles.parse_user_list(userFile)
        return

    # otherwise we have to make one for the user - check if one already there.
    if os.path.isfile(os.path.join("serverConfiguration", "userListRaw.csv")):
        print(
            "File 'userListRaw.csv' already exists in 'serverConfiguration'. Remove before continuing. Aborting."
        )
        exit(1)

    if demo:
        print(
            "Creating a demo user list at userListRaw.csv. ** DO NOT USE ON REAL SERVER **"
        )
        from plom.server import manageUserFiles

        rawfile = Path("serverConfiguration") / "userListRaw.csv"
        cl = pkg_resources.resource_string("plom", "demoUserList.csv")
        with open(rawfile, "wb") as fh:
            fh.write(cl)
        manageUserFiles.parse_user_list(rawfile)
        return

    if auto is not None:
        print("Creating an auto-generated user list at userListRaw.csv.")
        print(
            "Please edit as you see fit and then rerun 'plom-server users serverConfiguration/userListRaw.csv'"
        )
        from plom.server import manageUserFiles

        # grab required users and regular users
        lst = manageUserFiles.build_canned_users(auto)
        with open(os.path.join("serverConfiguration", "userListRaw.csv"), "w+") as fh:
            fh.write("user, password\n")
            for np in lst:
                fh.write('"{}", "{}"\n'.format(np[0], np[1]))

        return

    if not userFile:
        print(
            "Creating 'serverConfiguration/userListRaw.csv' - please edit passwords for 'manager', 'scanner', 'reviewer', and then add one or more normal users and their passwords. Note that passwords must be at least 4 characters and usernames should be at least 4 alphanumeric characters."
        )
        cl = pkg_resources.resource_string("plom", "templateUserList.csv")
        with open(os.path.join("serverConfiguration", "userListRaw.csv"), "wb") as fh:
            fh.write(cl)


#################
def checkDirectories():
    lst = [
        "pages",
        "pages/discardedPages",
        "pages/collidingPages",
        "pages/unknownPages",
        "pages/originalPages",
        "markedQuestions",
        "markedQuestions/plomFiles",
        "markedQuestions/commentFiles",
        "serverConfiguration",
    ]
    for d in lst:
        if not os.path.isdir(d):
            print(
                "Required directories are not present. Have you run 'plom-server init'?"
            )
            exit(1)


def checkServerConfigured():
    if not os.path.isfile(os.path.join("serverConfiguration", "serverDetails.toml")):
        print("Server configuration file not present. Have you run 'plom-server init'?")
        exit(1)

    if not os.path.isfile(os.path.join("serverConfiguration", "userList.json")):
        print("Processed userlist is not present. Have you run 'plom-server users'?")
        exit(1)

    if not (
        os.path.isfile(os.path.join("serverConfiguration", "plom.key"))
        and os.path.isfile(os.path.join("serverConfiguration", "plom-selfsigned.crt"))
    ):
        print("SSL keys not present. Have you run 'plom-server init'?")
        exit(1)

    if os.path.isfile(Path(specdir) / "predictionlist.csv"):
        print("Predictionlist present.")
    else:
        print(
            "Cannot find the predictionlist. Have you run 'plom-server init' yet? Aborting."
        )
        exit(1)


def prelaunchChecks():
    # check database, spec and classlist in place
    checkSpecAndDatabase()
    # check all directories built
    checkDirectories()
    # check serverConf and userlist present (also check predictionlist).
    checkServerConfigured()
    # ready to go
    return True


def launchTheServer(masterToken):
    from plom.server import theServer

    if prelaunchChecks():
        theServer.launch(masterToken)


#################

parser = argparse.ArgumentParser(
    epilog="Use '%(prog)s <subcommand> -h' for detailed help.\n\n"
    + server_instructions,
    formatter_class=argparse.RawDescriptionHelpFormatter,
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)
sub = parser.add_subparsers(
    dest="command", description="Perform various server-related tasks."
)
#
spI = sub.add_parser("init", help="Initialise server.")
spU = sub.add_parser("users", help="Create required users.")
spR = sub.add_parser("launch", help="Launch server.")
spR.add_argument(
    "masterToken",
    nargs="?",
    help="The master token is a 32 hex-digit string used to encrypt tokens in database. If you do not supply one then the server will create one. You should record the token somewhere (and reuse it at next server-start) if you want to be able to hot-restart the server (ie - restart the server without requiring users to log-off and log-in again).",
)
#
spU.add_argument(
    "userlist",
    nargs="?",
    help="Process the given userlist file OR if none given then produce a template.",
)
grp = spU.add_mutually_exclusive_group()
grp.add_argument(
    "--demo",
    action="store_true",
    help="Use demo auto-generated userlist and passwords. **DO NOT DO THIS ON REAL SERVER**",
)
grp.add_argument(
    "--auto",
    type=int,
    metavar="N",
    help="Construct an auto-generated user list of N users.",
)


def main():
    args = parser.parse_args()

    if args.command == "init":
        initialiseServer()
    elif args.command == "users":
        # process the class list and copy into place
        processUsers(args.userlist, args.demo, args.auto)
    elif args.command == "launch":
        launchTheServer(args.masterToken)
    else:
        parser.print_help()

    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian
# Copyright (C) 2020 Colin B. Macdonald

"""
Executable file frontend to the actual ID reader code.

Note: Code in this file is very similar to runTheReader code for the Sklearn
model.
"""

__copyright__ = "Copyright (C) 2018-20120 Andrew Rechnitzer and others"
__credits__ = ["Andrew Rechnitzer", "Dryden Wiebe", "Vala Vakilian"]
__license__ = "AGPLv3"

import json
import os
import sys

from .idReader import run_id_reader


if __name__ == "__main__":
    lock_file = sys.argv[1]

    if not os.path.isfile(lock_file):
        raise RuntimeError('Cannot acquire file "{}"'.format(lock_file))

    with open(lock_file) as fh:
        fileDictAndRect = json.load(fh)
        run_id_reader(fileDictAndRect[0], fileDictAndRect[1])

    os.unlink(lock_file)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

from datetime import datetime
import logging

from peewee import fn
import peewee as pw

from plom.rules import censorStudentNumber as censorID
from plom.rules import censorStudentName as censorName
from plom.db.tables import *

log = logging.getLogger("DB")


########## Bundle creation ##########


def createReplacementBundle(self):
    try:
        bref = Bundle.create(name="replacements")
    except IntegrityError as e:
        log.error("Create replacement page bundle = {}.{} error - {}".format(e))
        return False
    return True


def doesBundleExist(self, bundle_name, md5):
    """
    Checks if bundle with name=bundle_name or md5sum=md5 exists.
    4 possibilities
    * neither = no matching bundle, return [False]
    * name but not md5 = return [True, 'name'] - user is trying to upload different bundles with same name.
    * md5 but not name = return [True, 'md5sum'] - user is trying to same bundle with different names.
    * both match = return [True, 'both'] - user could be retrying after
      network failure (for example) or uploading unknown or colliding
      pages.
    """
    # check if that bundle-name is on file, and if the md5sum is known.
    bref = Bundle.get_or_none(name=bundle_name)
    if bref is not None:
        if bref.md5sum == md5:
            # name and md5sum match one in system
            # probably a crash occurred, so also return all the bundle_orders known in this bundle
            skip_list = []
            for iref in bref.images:
                skip_list.append(iref.bundle_order)
            return [True, "both", skip_list]

        else:
            return [True, "name"]
    # name not known, so just check md5sum
    if Bundle.get_or_none(md5sum=md5) is not None:
        return [True, "md5sum"]
    return [False, "no such bundle"]


def createNewBundle(self, bundle_name, md5):
    """
    Checks to see if bundle exists using 'doesBundleExist'.
    If bundle matches either 'name' or 'md5sum' then return [False, reason] - this shouldnt happen if scanner working correctly.
    If bundle matches 'both' then return [True, skip_list] where skip_list = the page-orders from that bundle that are already in the system. The scan scripts will then skip those uploads.
    If no such bundle return [True, []] - create the bundle and return an empty skip-list.
    """
    # use the doesBundleExist command logic to sanity check
    bundle_check = self.doesBundleExist(bundle_name, md5)
    if bundle_check[0] is False:
        Bundle.create(name=bundle_name, md5sum=md5)
        return [True, []]
    elif bundle_check[1] == "both":  # return [True, skip-list]
        bref = Bundle.get_or_none(name=bundle_name, md5sum=md5)
        skip_list = []
        for iref in bref.images:
            skip_list.append(iref.bundle_order)
        return [True, skip_list]
    else:
        return [False, bundle_check[1]]


########## Test creation stuff ##############
def areAnyPapersProduced(self):
    """True if any papers have been produced."""
    return len(Test.select()) > 0


def nextqueue_position(self):
    lastPos = Group.select(fn.MAX(Group.queue_position)).scalar()
    if lastPos is None:
        return 0
    else:
        return lastPos + 1


def createTest(self, t):
    with plomdb.atomic():
        try:
            tref = Test.create(test_number=t)  # must be unique
        except pw.IntegrityError as e:
            log.error("Create test {} error - {}".format(t, e))
            return False
    return True


def addTPages(self, tref, gref, t, pages, v):
    """
    For initial construction of test-pages for a test. We use these so we know what structured pages we should have.
    """
    flag = True
    with plomdb.atomic():
        for p in pages:
            try:
                TPage.create(
                    test=tref,
                    group=gref,
                    page_number=p,
                    version=v,
                    scanned=False,
                )
            except pw.IntegrityError as e:
                log.error("Adding page {} for test {} error - {}".format(p, t, e))
                flag = False
    return flag


def createIDGroup(self, t, pages):
    tref = Test.get_or_none(test_number=t)
    if tref is None:
        log.warning("Create IDGroup - No test with number {}".format(t))
        return False
    with plomdb.atomic():
        # make the Group
        gid = "i{}".format(str(t).zfill(4))
        try:
            gref = Group.create(
                test=tref,
                gid=gid,
                group_type="i",
                queue_position=self.nextqueue_position(),
            )  # must be unique
        except pw.IntegrityError as e:
            log.error(
                "Create ID - cannot create Group {} of test {} error - {}".format(
                    gid, t, e
                )
            )
            return False
        # make the IDGroup
        try:
            iref = IDGroup.create(test=tref, group=gref)
        except pw.IntegrityError as e:
            log.error(
                "Create ID - cannot create IDGroup {} of group {} error - {}.".format(
                    qref, gref, e
                )
            )
            return False
        return self.addTPages(tref, gref, t, pages, 1)  # always version 1.


def createDNMGroup(self, t, pages):
    tref = Test.get_or_none(test_number=t)
    if tref is None:
        log.warning("Create DNM - No test with number {}".format(t))
        return False

    gid = "d{}".format(str(t).zfill(4))
    with plomdb.atomic():
        # make the dnmgroup
        try:
            # A DNM group may have 0 pages, in that case mark it as scanned and set status = "complete"
            sc = True if len(pages) == 0 else False
            gref = Group.create(
                test=tref,
                gid=gid,
                group_type="d",
                scanned=sc,
                queue_position=self.nextqueue_position(),
            )
        except pw.IntegrityError as e:
            log.error(
                "Create DNM - cannot make Group {} of Test {} error - {}".format(
                    gid, t, e
                )
            )
            return False
        try:
            dref = DNMGroup.create(test=tref, group=gref)
        except pw.IntegrityError as e:
            log.error(
                "Create DNM - cannot create DNMGroup {} of group {} error - {}.".format(
                    dref, gref, e
                )
            )
            return False
        return self.addTPages(tref, gref, t, pages, 1)


def createQGroup(self, t, q, v, pages):
    tref = Test.get_or_none(test_number=t)
    if tref is None:
        log.warning("Create Q - No test with number {}".format(t))
        return False

    gid = "q{}g{}".format(str(t).zfill(4), q)

    with plomdb.atomic():
        # make the qgroup
        try:
            gref = Group.create(
                test=tref,
                gid=gid,
                group_type="q",
                version=v,
                queue_position=self.nextqueue_position(),
            )
        except pw.IntegrityError as e:
            log.error(
                "Create Q - cannot create group {} of Test {} error - {}".format(
                    gid, t, e
                )
            )
            return False
        try:
            qref = QGroup.create(test=tref, group=gref, question=q, version=v)
        except pw.IntegrityError as e:
            log.error(
                "Create Q - cannot create QGroup of question {} error - {}.".format(
                    gid, e
                )
            )
            return False
        ## create annotation 0 owned by HAL
        try:
            uref = User.get(name="HAL")
            aref = Annotation.create(qgroup=qref, edition=0, user=uref)
        except pw.IntegrityError as e:
            log.error(
                "Create Q - cannot create Annotation  of question {} error - {}.".format(
                    gid, e
                )
            )
            return False

        return self.addTPages(tref, gref, t, pages, v)


def getPageVersions(self, t):
    """Get the mapping between page numbers and version for a test.

    Args:
        t (int): a paper number.

    Returns:
        dict: keys are page numbers (int) and value is the page
            version (int), or empty dict if there was no such paper.
    """
    tref = Test.get_or_none(test_number=t)
    if tref is None:
        return {}
    else:
        pvDict = {p.page_number: p.version for p in tref.tpages}
        return pvDict


def produceTest(self, t):
    """Someone has told us they produced (made PDF) for this paper.

    Args:
        t (int): a paper number.

    Exceptions:
        IndexError: no such paper exists.
        ValueError: you already told us you made it.
    """
    tref = Test.get_or_none(test_number=t)
    if tref is None:
        log.error('Cannot set paper {} to "produced" - it does not exist'.format(t))
        raise IndexError("Paper number does not exist: out of range?")
    else:
        # TODO - work out how to make this more efficient? Multiple updates in one op?
        with plomdb.atomic():
            if tref.produced:
                # TODO be less harsh if we have the same md5sum
                log.error('Paper {} was already "produced"!'.format(t))
                raise ValueError("Paper was already produced")
            tref.produced = True
            tref.save()
        log.info('Paper {} is set to "produced"'.format(t))


def id_paper(self, paper_num, user_name, sid, sname):
    """Associate student name and id with a paper in the database.

    See also :func:`plom.db.db_identify.ID_id_paper` which is similar.

    Args:
        paper_num (int)
        user_name (str): User who did the IDing.
        sid (str): student id.
        sname (str): student name.

    Returns:
        tuple: `(True, None, None)` if succesful, `(False, 409, msg)`
            means `sid` is in use elsewhere, a serious problem for
            the caller to deal with.  `(False, int, msg)` covers all
            other errors.  `msg` gives details about errors.  Some
            of these should not occur, and indicate possible bugs.
            `int` gives a hint of suggested HTTP status code,
            currently it can be 404 or 409.

    TODO: perhaps several sorts of exceptions would be better.
    """
    uref = User.get(name=user_name)  # TODO: or hardcode HAL like before
    # since user authenticated, this will always return legit ref.

    logbase = 'User "{}" tried to ID paper {}'.format(user_name, paper_num)
    with plomdb.atomic():
        tref = Test.get_or_none(Test.test_number == paper_num)
        if tref is None:
            msg = "denied b/c paper not found"
            log.error("{}: {}".format(logbase, msg))
            return False, 404, msg
        iref = tref.idgroups[0]
        iref.user = uref
        iref.status = "done"
        iref.student_id = sid
        iref.student_name = sname
        iref.identified = True
        iref.time = datetime.now()
        try:
            iref.save()
        except pw.IntegrityError:
            msg = "student id {} already entered elsewhere".format(censorID(sid))
            log.error("{} but {}".format(logbase, msg))
            return False, 409, msg
        tref.identified = True
        tref.save()
        log.info(
            'Paper {} ID\'d by "{}" as "{}" "{}"'.format(
                paper_num, user_name, censorID(sid), censorName(sname)
            )
        )
    return True, None, None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '../qtCreatorFiles/ui_test_view.ui'
#
# Created by: PyQt5 UI code generator 5.12.3
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_TestView(object):
    def setupUi(self, TestView):
        TestView.setObjectName("TestView")
        TestView.resize(714, 603)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(TestView)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.frame_2 = QtWidgets.QFrame(TestView)
        self.frame_2.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame_2.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame_2.setObjectName("frame_2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.frame_2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.groupViewTabWidget = QtWidgets.QTabWidget(self.frame_2)
        self.groupViewTabWidget.setObjectName("groupViewTabWidget")
        self.verticalLayout.addWidget(self.groupViewTabWidget)
        self.verticalLayout_2.addWidget(self.frame_2)
        self.frame = QtWidgets.QFrame(TestView)
        self.frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.frame.setObjectName("frame")
        self.gridLayout = QtWidgets.QGridLayout(self.frame)
        self.gridLayout.setObjectName("gridLayout")
        self.prevGroupButton = QtWidgets.QPushButton(self.frame)
        self.prevGroupButton.setObjectName("prevGroupButton")
        self.gridLayout.addWidget(self.prevGroupButton, 0, 0, 1, 1)
        self.nextGroupButton = QtWidgets.QPushButton(self.frame)
        self.nextGroupButton.setObjectName("nextGroupButton")
        self.gridLayout.addWidget(self.nextGroupButton, 0, 1, 1, 1)
        self.closeButton = QtWidgets.QPushButton(self.frame)
        self.closeButton.setObjectName("closeButton")
        self.gridLayout.addWidget(self.closeButton, 1, 3, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 1, 2, 1, 1)
        self.maxNormButton = QtWidgets.QPushButton(self.frame)
        self.maxNormButton.setObjectName("maxNormButton")
        self.gridLayout.addWidget(self.maxNormButton, 1, 1, 1, 1)
        self.verticalLayout_2.addWidget(self.frame)

        self.retranslateUi(TestView)
        self.groupViewTabWidget.setCurrentIndex(-1)
        QtCore.QMetaObject.connectSlotsByName(TestView)

    def retranslateUi(self, TestView):
        _translate = QtCore.QCoreApplication.translate
        TestView.setWindowTitle(_translate("TestView", "View test pages"))
        self.prevGroupButton.setText(_translate("TestView", "Previous Page"))
        self.nextGroupButton.setText(_translate("TestView", "Next Page"))
        self.closeButton.setText(_translate("TestView", "Close"))
        self.maxNormButton.setText(_translate("TestView", "Toggle Maximise"))



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Vala Vakilian

from aiohttp import web, MultipartWriter, MultipartReader

from .routeutils import authenticate_by_token, authenticate_by_token_required_fields


class ReportHandler:
    def __init__(self, plomServer):
        self.server = plomServer

    # @routes.get("/REP/scanned")
    @authenticate_by_token_required_fields(["user"])
    def RgetScannedTests(self, data, request):
        """Respond with a dictionary of completed exams.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/scanned.

        Returns:
            aiohttp.web_response.Response: [description]
        """

        if not data["user"] in ("manager", "scanner"):
            return web.Response(status=401)
        # A dictionary that involves the complete tests : key:exam_number, value: list of lists
        #   involving the page and the page version.
        # Ex: {2: [['t.1', 1], ['t.2', 1], ['t.3', 1], ['t.4', 1], ['t.5', 1], ['t.6', 1]], ... }
        return web.json_response(self.server.RgetScannedTests(), status=200)

    # @routes.get("/REP/incomplete")
    @authenticate_by_token_required_fields(["user"])
    def RgetIncompleteTests(self, data, request):
        """Respond with the incomplete exams, providing information on individual pages.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/incomplete.

        Returns:
            aiohttp.web_response.Response: A response which includes a dictionary of pages for
                incomplete exams.
        """

        if not data["user"] in ("manager", "scanner"):
            return web.Response(status=401)
        # The response is a dictionary of the form:
        # {test_number: [[test string, TODO: Version number ?, True/False for page incomplete or not], ...], ...}
        # Ex: {1: [['t.1', 1, True], ['t.2', 1, True], ['t.3', 2, True], ['t.4', 1, True], ['t.5', 2, True], ['t.6', 2, False]]}
        return web.json_response(self.server.RgetIncompleteTests(), status=200)

    # @routes.get("/REP/completeHW")
    @authenticate_by_token_required_fields(["user"])
    def RgetCompleteHW(self, d, request):
        if not d["user"] in ("manager", "scanner"):
            return web.Response(status=401)
        return web.json_response(self.server.RgetCompleteHW(), status=200)

    # @routes.get("/REP/missingHW")
    @authenticate_by_token_required_fields(["user"])
    def RgetMissingHWQ(self, d, request):
        if not d["user"] in ("manager", "scanner"):
            return web.Response(status=401)
        return web.json_response(self.server.RgetMissingHWQ(), status=200)

    # @routes.get("/REP/unused")
    @authenticate_by_token_required_fields(["user"])
    def RgetUnusedTests(self, d, request):
        # TODO: Requires documentation.
        if not d["user"] in ("manager", "scanner"):
            return web.Response(status=401)
        return web.json_response(self.server.RgetUnusedTests(), status=200)

    # @routes.get("/REP/progress")
    @authenticate_by_token_required_fields(["user", "q", "v"])
    def RgetProgress(self, data, request):
        """Respond with an overall progress status of the marking process.

        Responds with status 200/401.

        Args:
            data (dict): Dictionary including user data in addition to question number and version.
            request (aiohttp.web_request.Request): Request of type GET /REP/progress.

        Returns:
            aiohttp.web_response.Response: A response which includes a dictionary of pages for
                incomplete exams.
        """

        # An example of the report sent as a response:  TODO: Should I explain this better.
        # {'NScanned': 10, 'NMarked': 7, 'NRecent': 7, 'avgMark': 4.428571428571429, 'avgMTime': 153.57142857142858}
        if not data["user"] == "manager":
            return web.Response(status=401)
        return web.json_response(
            self.server.RgetProgress(data["q"], data["v"]), status=200
        )

    # @routes.get("/REP/questionUserProgress")
    @authenticate_by_token_required_fields(["user", "q", "v"])
    def RgetQuestionUserProgress(self, data, request):
        """Respond with information on each user's progress on grading of a question version.

        Responds with status 200/401.

        Args:
            data (dict): Dictionary including user data in addition to question number and version.
            request (aiohttp.web_request.Request): Request GET /REP/questionUserProgress.

        Returns:
            aiohttp.web_response.Response: A response with information on the progress of each
                use on each question.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)

        # A list of list with the following response:
        # [ Number of scanned papers, [Username, Number of marked], [Username, Number of marked], etc]
        return web.json_response(
            self.server.RgetQuestionUserProgress(data["q"], data["v"]), status=200
        )

    # @routes.get("/REP/markHistogram")
    @authenticate_by_token_required_fields(["user", "q", "v"])
    def RgetMarkHistogram(self, data, request):
        """Returns histogram info for the grading of a question.

        Responds with status 200/401.

        Args:
            data (dict): Dictionary including user data in addition to question number.
            request (aiohttp.web_request.Request): Request of type GET /REP/markHistogram.

        Returns:
            aiohttp.web_response.Response: A response object with the grading histogram info
                for a question.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        # Respond with the corresponding values for non-zero bars in the histogram.
        # A dictionary with usernames as keys and the bar values for non zero bars as dictionaries.
        # TODO: What did the columns represent ?
        # Example: {'user0': {5: 3, 4: 1}}
        return web.json_response(
            self.server.RgetMarkHistogram(data["q"], data["v"]), status=200
        )

    # @routes.get("/REP/progress")
    @authenticate_by_token_required_fields(["user"])
    def RgetIdentified(self, data, request):
        """Respond with a dictionary of identified papers

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): GET /REP/identified request type.

        Returns:
            aiohttp.web_response.Response: A response object including a dictionary of
                identified papers.

        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        return web.json_response(self.server.RgetIdentified(), status=200)

    # @routes.get("/REP/completionStatus")
    @authenticate_by_token_required_fields(["user"])
    def RgetCompletionStatus(self, data, request):
        """Respond with a status of the complete papers providing information on grading progress.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/completionStatus.

        Returns:
            aiohttp.web_response.Response: A response which includes a dictionary of the grading
                status.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        # A dictionary which includes info about the grading status:
        # {exam number: [Test identifies, Total score calculated, number of marked pages], ...}
        # {2: [True, True, 0], ...}
        return web.json_response(self.server.RgetCompletionStatus(), status=200)

    # @routes.get("/REP/outToDo")
    @authenticate_by_token_required_fields(["user"])
    def RgetOutToDo(self, data, request):
        """Respond with a list of tasks that are currently out with clients.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): GET /REP/outToDo type request.

        Returns:
            aiohttp.web_response.Response: A response that includes a list of todo
                tasks.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)

        # Response includes a list of [task code, username, time of task sent out].
        # Ex: [['mrk-t11-q1-v1', 'user0', '20:06:21-01:21:47'], ... ]
        return web.json_response(self.server.RgetOutToDo(), status=200)

    # @routes.get("/REP/marked")
    @authenticate_by_token_required_fields(["user", "q", "v"])
    def RgetMarked(self, d, request):
        # TODO: Requires documentation.
        if not d["user"] == "manager":
            return web.Response(status=401)
        return web.json_response(self.server.RgetMarked(d["q"], d["v"]), status=200)

    # @routes.get("/REP/status/{test}")
    @authenticate_by_token_required_fields(["user"])
    def RgetStatus(self, data, request):
        """Respond with the marking status of an exam.

        Responds with status 200/401/404.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of typeGET /REP/status/2 which also
                includes the test number.

        Returns:
            aiohttp.web_response.Response: A response including a dictionary for information on
                grading status for an exam.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        test_number = request.match_info["test"]
        marking_status_response = self.server.RgetStatus(test_number)
        status_response_success = marking_status_response[0]

        # An example of gradding status summary can be seen below:
        #  {'number': 2, 'identified': True, 'marked': False, 'totalled': False, 'sid': '10130103', 'sname': 'Vandeventer, Irene',
        # 'iwho': 'HAL', 1: {'marked': False, 'version': 2}, 2: {'marked': False, 'version': 1}, 3: {'marked': False, 'version': 2}}
        # TODO: Whats is iwho ?
        if status_response_success:
            marking_status_dict = marking_status_response[1]
            return web.json_response(marking_status_dict, status=200)
        else:
            return web.Response(status=404)

    # @routes.get("/REP/spreadSheet")
    @authenticate_by_token_required_fields(["user"])
    def RgetSpreadsheet(self, data, request):
        """Creates a spreadsheet csv file of the grading information.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/spreadSheet.

        Returns:
            aiohttp.web_response.Response: A response object that includes the csv dictionary
                for the grading results spreadsheet.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        rmsg = self.server.RgetSpreadsheet()

        # TODO: Understand this better.
        # For each paper, the response includes the following information:
        # {1: {'identified': True, 'marked': False, 'totalled': False, 'sid': '10050380',
        # 'sname': 'Fink, Iris', 'q1v': 2, 'q1m': '', 'q2v': 1, 'q2m': '', 'q3v': 2, 'q3m': ''}
        return web.json_response(rmsg, status=200)

    # @routes.get("/REP/coverPageInfo/{test}")
    @authenticate_by_token_required_fields(["user"])
    def RgetCoverPageInfo(self, d, request):
        # TODO: Requires documentation.
        if not d["user"] == "manager":
            return web.Response(status=401)
        testNumber = request.match_info["test"]
        rmsg = self.server.RgetCoverPageInfo(testNumber)
        return web.json_response(rmsg, status=200)

    # @routes.get("/REP/originalFiles/{test}")
    @authenticate_by_token_required_fields(["user"])
    def RgetOriginalFiles(self, d, request):
        # TODO: Requires documentation.
        if not d["user"] == "manager":
            return web.Response(status=401)
        testNumber = request.match_info["test"]
        rmsg = self.server.RgetOriginalFiles(testNumber)

        if len(rmsg) > 0:
            return web.json_response(rmsg, status=200)
        else:
            return web.Response(status=404)

    # @routes.get("/REP/annotatedFiles/{test}")
    @authenticate_by_token_required_fields(["user"])
    def RgetAnnotatedFiles(self, d, request):
        # TODO: Requires documentation.
        if not d["user"] == "manager":
            return web.Response(status=401)
        testNumber = request.match_info["test"]
        rmsg = self.server.RgetAnnotatedFiles(testNumber)

        if len(rmsg) > 0:
            return web.json_response(rmsg, status=200)
        else:
            return web.Response(status=404)

    # @routes.get("/REP/userList")
    @authenticate_by_token_required_fields(["user"])
    def RgetUserList(self, data, request):
        """Return a list of Plom users.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/userList.

        Returns:
            aiohttp.web_response.Response: A response object entailing a list of Plom users.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)

        return web.json_response(self.server.RgetUserList(), status=200)

    # @routes.get("/REP/userDetails")
    @authenticate_by_token_required_fields(["user"])
    def RgetUserDetails(self, data, request):
        """Gets a list of users and their detail.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request GET /REP/userDetails.

        Returns:
            aiohttp.web_response.Response: A response object entailing a list of plom users.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)

        # A list of list for each user involving the following format:
        # [ Username, User enabled, User logged in, Last activity time, Last action,
        #   Papers ID'd by user, Papers totalled by user, Question's marked by user]
        return web.json_response(self.server.RgetUserDetails(), status=200)

    # @routes.get("/REP/markReview")
    @authenticate_by_token_required_fields(["user", "filterQ", "filterV", "filterU"])
    def RgetMarkReview(self, data, request):
        """Respond with a list of graded tasks that match the filter description.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary which includes the user data in addition to the
                filter query information sent by the client.
            request (aiohttp.web_request.Request): Request of type GET /REP/markReview.

        Returns:
            aiohttp.web_response.Response: Response object including  metadata on the
                graded exams that match the filter.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        rmsg = self.server.RgetMarkReview(
            data["filterQ"], data["filterV"], data["filterU"]
        )

        # A list of lists including metadata information for the graded exams mathing the filter with the format of:
        # [Test number, Question number, Version number, Mark, Username, # TODO: Explain Marking Time, date/time of marking ]
        # Ex:[[3, 1, 1, 5, 'user0', 7, '20:06:21-01:21:56'],
        return web.json_response(rmsg, status=200)

    # @routes.get("/REP/idReview")
    @authenticate_by_token_required_fields(["user"])
    def RgetIDReview(self, data, request):
        """Respond with metadata about identified papers.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/idReview.

        Returns:
            aiohttp.web_response.Response: A response including metadata about the identified
                papers queued for reviewing.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        rmsg = self.server.RgetIDReview()

        # A list of lists included information of format below:
        # [Test number, User who ID'd the paper, Time of ID'ing, Student ID, Student name]
        return web.json_response(rmsg, status=200)

    # @routes.get("/REP/totReview")
    @authenticate_by_token_required_fields(["user"])
    def RgetTotReview(self, data, request):
        """Return information about papers which have total grades calculated.

        Responds with status 200/401.

        Args:
            data (dict): A dictionary having the user/token.
            request (aiohttp.web_request.Request): Request of type GET /REP/totReview.

        Returns:
            aiohttp.web_response.Response: A response including metadata about the totalled
                papers.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        rmsg = self.server.RgetTotReview()

        # Returns a list of lists with the format of:
        # [Test number, User who did the totalling, Time of totalling, Total mark]
        return web.json_response(rmsg, status=200)

    # @routes.get("/REP/annotatedImage")
    @authenticate_by_token_required_fields(
        ["user", "testNumber", "questionNumber", "version"]
    )
    def RgetAnnotatedImage(self, data, request):
        """Responses with the annotated image of the requested question.

        Responds with status 200/404.

        Args:
            data (dict): Dictionary including user data in addition to question number.
            request (aiohttp.web_request.Request): Request of type GET /REP/annotatedImage.

        Returns:
            aiohttp.web_response.Response: A response including the path to the annotated image.
        """

        if not data["user"] == "manager":
            return web.Response(status=401)
        annotated_page_response = self.server.RgetAnnotatedImage(
            data["testNumber"], data["questionNumber"], data["version"]
        )

        retrieve_annotated_success = annotated_page_response[0]

        if retrieve_annotated_success:
            # TODO: Shouldn't this be annotated_page_response[1:] ?
            annotated_image_paths = annotated_page_response[1]

            return web.FileResponse(annotated_page_response[1], status=200)
        else:
            return web.Response(status=404)

    def setUpRoutes(self, router):
        """Adds the response functions to the router object.

        Args:
            router (aiohttp.web_urldispatcher.UrlDispatcher): Router object which we will add the
                response functions to.
        """

        router.add_get("/REP/scanned", self.RgetScannedTests)
        router.add_get("/REP/incomplete", self.RgetIncompleteTests)
        router.add_get("/REP/completeHW", self.RgetCompleteHW)
        router.add_get("/REP/missingHW", self.RgetMissingHWQ)
        router.add_get("/REP/unused", self.RgetUnusedTests)
        router.add_get("/REP/progress", self.RgetProgress)
        router.add_get("/REP/questionUserProgress", self.RgetQuestionUserProgress)
        router.add_get("/REP/markHistogram", self.RgetMarkHistogram)
        router.add_get("/REP/marked", self.RgetMarked)
        router.add_get("/REP/identified", self.RgetIdentified)
        router.add_get("/REP/completionStatus", self.RgetCompletionStatus)
        router.add_get("/REP/outToDo", self.RgetOutToDo)
        router.add_get("/REP/status/{test}", self.RgetStatus)
        router.add_get("/REP/spreadSheet", self.RgetSpreadsheet)
        router.add_get("/REP/originalFiles/{test}", self.RgetOriginalFiles)
        router.add_get("/REP/coverPageInfo/{test}", self.RgetCoverPageInfo)
        router.add_get("/REP/annotatedFiles/{test}", self.RgetAnnotatedFiles)
        router.add_get("/REP/userList", self.RgetUserList)
        router.add_get("/REP/userDetails", self.RgetUserDetails)
        router.add_get("/REP/markReview", self.RgetMarkReview)
        router.add_get("/REP/idReview", self.RgetIDReview)
        router.add_get("/REP/totReview", self.RgetTotReview)
        router.add_get("/REP/annotatedImage", self.RgetAnnotatedImage)



# -*- coding: utf-8 -*-

"""Check which students have submitted what in the submittedHWByQ and submittedHWExtra directories"""

__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer and Colin B. Macdonald"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"
# SPDX-License-Identifier: AGPL-3.0-or-later

from collections import defaultdict
import glob
import getpass
import os

from plom.rules import isValidStudentNumber
from plom.messenger import ScanMessenger
from plom.plom_exceptions import *


def IDQorIDorBad(fullfname):
    """Factor filename into one of two forms or answer name is bad.

    Args:
        fullfname (str, Pathlib): a filename that is supposed to have
            a particular form.

    Returns:
        list: first entry is either "IDQ" or "JID" or "BAD", with other
            entries following in the "IDQ" and "JID" cases.
    """
    fname = os.path.basename(fullfname)
    splut = fname.split(".")
    try:
        QFlag = splut[-2].isnumeric() or splut[-2] == "_"
        IDFlag = isValidStudentNumber(splut[-3])
    except IndexError:
        return ["BAD"]
    if QFlag and IDFlag:  # [-3] is ID and [-2] is Q.
        return ["IDQ", splut[-3], splut[-2]]  # ID and Q
    elif isValidStudentNumber(splut[-2]):  # [-2] is ID
        return ["JID", splut[-2]]  # Just ID
    else:
        return ["BAD"]  # Bad format


def whoSubmittedWhatOnDisc():
    print(">> Checking submissions in local 'submittedHWByQ' subdirectory <<")
    hwByQ = defaultdict(list)
    hwOne = defaultdict(list)
    problemFQ = []
    problemOF = []

    for fn in glob.glob(os.path.join("submittedHWByQ", "*.pdf")):
        IDQ = IDQorIDorBad(fn)
        if len(IDQ) == 3:
            sid, q = IDQ[1:]
            hwByQ[sid].append([fn, q])
        else:
            # print("File {} has incorrect format for homework-by-question".format(fn))
            problemFQ.append(os.path.basename(fn))

    for fn in glob.glob(os.path.join("submittedLoose", "*.pdf")):
        IDQ = IDQorIDorBad(fn)
        if len(IDQ) == 2:
            sid = IDQ[1]
            hwOne[sid].append(fn)
        else:
            # print("File {} has incorrect format for homework-by-question".format(fn))
            problemOF.append(os.path.basename(fn))

    for sid in sorted(hwByQ.keys()):
        print("#{} submitted q's {}".format(sid, sorted([x[1] for x in hwByQ[sid]])))

    for sid in sorted(hwOne.keys()):
        print("#{} submitted loose pages".format(sid))

    warn = []
    for sid in sorted(hwOne.keys()):
        if sid in hwByQ:
            warn.append(sid)
    if len(warn) > 0:
        print(">>> Warning <<<")
        print(
            "These students submitted both HW by Q, and HW in loose pages: {}".format(
                warn
            )
        )
    if len(problemFQ) > 0:
        print(">>> Warning <<<")
        print(
            "These files in submittedHWByQ have the wrong name format: {}".format(
                problemFQ
            )
        )
        print("Please check them before proceeding. They will not be processed.")
    if len(problemOF) > 0:
        print(">>> Warning <<<")
        print(
            "These files in submittedHWExtra have the wrong name format: {}".format(
                problemOF
            )
        )
        print("Please check them before proceeding. They will not be processed.")


def whoSubmittedWhatOnServer(server, password):
    if server and ":" in server:
        s, p = server.split(":")
        msgr = ScanMessenger(s, port=p)
    else:
        msgr = ScanMessenger(server)
    msgr.start()

    # get the password if not specified
    if password is None:
        try:
            pwd = getpass.getpass("Please enter the 'scanner' password:")
        except Exception as error:
            print("ERROR", error)
    else:
        pwd = password

    # get started
    try:
        msgr.requestAndSaveToken("scanner", pwd)
    except PlomExistingLoginException:
        print(
            "You appear to be already logged in!\n\n"
            "  * Perhaps a previous session crashed?\n"
            "  * Do you have another scanner-script running,\n"
            "    e.g., on another computer?\n\n"
            'In order to force-logout the existing authorisation run "plom-hwscan clear"'
        )
        exit(10)

    missingHWQ = msgr.getMissingHW()  # passes back dict
    completeHW = msgr.getCompleteHW()  # passes back list [test_number, sid]

    msgr.closeUser()
    msgr.stop()

    print(">> Checking incomplete submissions on server <<")
    print("The following students have complete submissions (each question present)")
    print(", ".join(sorted([x[1] for x in completeHW])))
    print(
        "The following students have incomplete submissions (missing questions indicated)"
    )
    for t in missingHWQ:
        print("{} missing {}".format(missingHWQ[t][1], missingHWQ[t][2:]))


def whoSubmittedWhat(server=None, password=None, directory_check=False):
    if directory_check:
        whoSubmittedWhatOnDisc()
    else:
        whoSubmittedWhatOnServer(server, password)


def verifiedComplete(server=None, password=None):

    if server and ":" in server:
        s, p = server.split(":")
        msgr = ScanMessenger(s, port=p)
    else:
        msgr = ScanMessenger(server)
    msgr.start()

    # get the password if not specified
    if password is None:
        try:
            pwd = getpass.getpass("Please enter the 'scanner' password:")
        except Exception as error:
            print("ERROR", error)
    else:
        pwd = password

    # get started
    try:
        msgr.requestAndSaveToken("scanner", pwd)
    except PlomExistingLoginException:
        print(
            "You appear to be already logged in!\n\n"
            "  * Perhaps a previous session crashed?\n"
            "  * Do you have another scanner-script running,\n"
            "    e.g., on another computer?\n\n"
            'In order to force-logout the existing authorisation run "plom-hwscan clear"'
        )
        exit(10)

    # grab number of questions - so we can work out what is missing
    spec = msgr.get_spec()
    numberOfQuestions = spec["numberOfQuestions"]

    msgr.closeUser()
    msgr.stop()

    hwByQ = defaultdict(list)
    for fn in glob.glob("submittedHWByQ/*.pdf"):
        IDQ = IDQorIDorBad(fn)
        if len(IDQ) == 3:
            sid, q = IDQ[1:]
            hwByQ[sid].append([fn, q])
    # return fileNames belonging to complete homeworks
    validFiles = []
    for sid in hwByQ:
        if len(hwByQ[sid]) == numberOfQuestions:
            validFiles += [x[0] for x in hwByQ[sid]]
    return validFiles



__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2018-2019 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPLv3"

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPainter, QPixmap
from PyQt5.QtWidgets import (
    QDialog,
    QGridLayout,
    QGraphicsPixmapItem,
    QGraphicsScene,
    QGraphicsView,
    QGraphicsWidget,
    QLabel,
    QPushButton,
    QSpinBox,
)


class ExamReorientWindow(QDialog):
    """Widget to flip page images if they happen to be upsidedown.
    User selects how many pages in the group, the image is then
    split into that number of sub-images. They can then be flipped
    independently. The result is then saved.
    """

    def __init__(self, fname):
        QGraphicsWidget.__init__(self)
        self.initUI(fname)

    def initUI(self, fname):
        # Grab the image filename.
        self.fname = fname
        # Set up a QGraphicsScene and QGraphicsView
        self.scene = QGraphicsScene()
        self.view = QGraphicsView(self.scene)
        # Create a pixmap of the file and corresponding item
        self.image = QPixmap(self.fname)
        self.imageItem = QGraphicsPixmapItem(self.image)
        self.imageItem.setTransformationMode(Qt.SmoothTransformation)
        # Set scene dimensions to that of the image.
        self.scene.setSceneRect(0, 0, self.image.width(), self.image.height())
        # Set the view to encompass the image.
        self.scene.addItem(self.imageItem)
        self.view.fitInView(self.imageItem, Qt.KeepAspectRatio)

        # Spinbox for entering the number of pages in the image
        self.partL = QLabel("Number of pages")
        self.partSB = QSpinBox()
        self.partSB.setValue(1)
        self.partSB.setRange(1, 9)

        # Buttons and connections to functions
        self.splitB = QPushButton("Split")
        self.splitB.clicked.connect(self.splitIt)
        self.acceptB = QPushButton("Accept")
        self.acceptB.clicked.connect(self.acceptIt)
        self.cancelB = QPushButton("Cancel")
        self.cancelB.clicked.connect(self.reject)
        # The void button will become a flip button for each page
        self.voidB = QPushButton()
        self.voidB.setEnabled(False)
        # lay out buttons etc
        self.grid = QGridLayout()
        self.grid.addWidget(self.partL, 1, 1)
        self.grid.addWidget(self.partSB, 1, 2)
        self.grid.addWidget(self.view, 1, 3, 3, 3)
        self.grid.addWidget(self.voidB, 5, 3, 1, 3)
        self.grid.addWidget(self.splitB, 2, 1, 2, 1)
        self.grid.addWidget(self.acceptB, 10, 2)
        self.grid.addWidget(self.cancelB, 10, 1)
        self.setLayout(self.grid)
        self.show()

    def splitIt(self):
        """Split the groupimage into user-selected number of pages
        and create individual page-flip buttons.
        """
        self.splitB.setEnabled(False)
        n = self.partSB.value()
        self.grid.removeWidget(self.view)
        self.grid.addWidget(self.view, 1, 3, 3, n)
        # remove the voidButton and replace with a button for each page.
        self.grid.removeWidget(self.voidB)
        self.flip = {}
        for k in range(n):
            self.flip[k] = QPushButton("{}".format(k))
            self.flip[k].clicked.connect(lambda: self.flipPage())
            self.grid.addWidget(self.flip[k], 5, 3 + k)
        # remove the original image and paste in each nth portion of it.
        self.scene.removeItem(self.imageItem)
        w = self.image.width() // n
        h = self.image.height()
        # Each split image and corresponding graphicsitem
        self.splitImg = {}
        self.splitImgI = {}
        for k in range(n):
            self.splitImg[k] = self.image.copy(w * k, 0, w, h)
            self.splitImgI[k] = QGraphicsPixmapItem(self.splitImg[k])
            self.splitImgI[k].setPos(w * k, 0)
            # Flip around centre.
            self.splitImgI[k].setTransformOriginPoint(w // 2, h // 2)
            # Place into the scene
            self.scene.addItem(self.splitImgI[k])
        self.view.fitInView(0, 0, w * n, h)

    def flipPage(self):
        # rotate the relevant nth of the groupimage by 180 around its centre
        sender = self.sender()
        k = int(sender.text().replace("&", ""))
        if self.splitImgI[k].rotation() == 0:
            self.splitImgI[k].setRotation(180)
        else:
            self.splitImgI[k].setRotation(0)

    def acceptIt(self):
        # Accept the result and save.
        w = self.image.width()
        h = self.image.height()
        # create pixmap of scene and painter to render it.
        oimg = QPixmap(w, h)
        exporter = QPainter(oimg)
        self.scene.render(exporter)
        exporter.end()
        # save result to file.
        oimg.save(self.fname)
        self.accept()



# -*- coding: utf-8 -*-

"""Tools for working with TeX"""

__copyright__ = "Copyright (C) 2018-2020 Andrew Rechnitzer and Colin B. Macdonald"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"
# SPDX-License-Identifier: AGPL-3.0-or-later

import os
import sys
import subprocess
import tempfile
import shutil
from pathlib import Path

import pkg_resources


def texFragmentToPNG(fragment, outName):
    """Process a fragment of latex and produce a png image."""

    head = r"""
    \documentclass[12pt]{article}
    \usepackage[letterpaper, textwidth=5in]{geometry}
    \usepackage{amsmath, amsfonts}
    \usepackage{xcolor}
    \usepackage[active, tightpage]{preview}
    \begin{document}
    \begin{preview}
    \color{red}
    """

    foot = r"""
    \end{preview}
    \end{document}
    """

    # make a temp dir to build latex in
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, "frag.tex"), "w") as fh:
            fh.write(head)
            fh.write(fragment)
            fh.write(foot)

        latexIt = subprocess.run(
            [
                "latexmk",
                "-quiet",
                "-interaction=nonstopmode",
                "-no-shell-escape",
                "frag.tex",
            ],
            cwd=tmpdir,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if latexIt.returncode != 0:
            return False

        convertIt = subprocess.run(
            [
                "dvipng",
                "-q",
                "-D",
                "225",
                "-bg",
                "transparent",
                "frag.dvi",
                "-o",
                "frag.png",
            ],
            cwd=tmpdir,
            stdout=subprocess.DEVNULL,
        )
        if convertIt.returncode != 0:
            # sys.exit(convertIt.returncode)
            return False

        shutil.copyfile(os.path.join(tmpdir, "frag.png"), outName)
    return True


def buildLaTeX(src, out):
    """Compile a string or bytes of latex.

    Args:
        src (str, bytes):
        out (file-like):

    Returns:
        exit value from the subprocess call (zero good, non-zero BAD)
        stdout/stderr from the subprocess call

    TODO: this is more generally useful but how to handle the idBox2?
    """

    with tempfile.TemporaryDirectory() as tmpdir:

        tmp = pkg_resources.resource_string("plom", "testTemplates/idBox2.pdf")
        with open(Path(tmpdir) / "idBox2.pdf", "wb") as fh:
            fh.write(tmp)

        # TODO: this is not very duck-type of us!
        if isinstance(src, bytes):
            mode = "wb"
        else:
            mode = "w"

        with open(Path(tmpdir) / "stuff.tex", mode) as fh:
            fh.write(src)

        latexIt = subprocess.run(
            [
                "latexmk",
                "-pdf",
                "-interaction=nonstopmode",
                "-no-shell-escape",
                "stuff.tex",
            ],
            cwd=tmpdir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if latexIt.returncode == 0:
            out.write((Path(tmpdir) / "stuff.pdf").read_bytes())

    return latexIt.returncode, latexIt.stdout.decode()



from pytest import raises

import plom.plom_exceptions
from plom.plom_exceptions import *


def test_plom_exc_string():
    e = PlomException("foo")
    assert str(e) == "foo"


def test_exc_inheritance():
    e = PlomAuthenticationException()
    assert isinstance(e, PlomException)


def test_exc_auth_has_default_msg():
    e = PlomAuthenticationException()
    assert "authenticate" in str(e).lower()
    e = PlomAuthenticationException("foo")
    assert str(e) == "foo"


def test_exc_all_print_properly():
    excs = [eval(e) for e in dir(plom.plom_exceptions) if e.startswith("Plom")]
    assert len(excs) > 10  # we have some exceptions
    for exc in excs:
        assert str(exc("foo")) == "foo"



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Colin B. Macdonald

import os
import subprocess

import jpegtran


def rotateBitmap(fname, angle):
    """Rotate bitmap, (almost) lossless for jpg.

    args:
        filename (str): name of a file
        angle (int): 0, 90, 180, 270, or -90 degree rotation.

    TODO: think about multiples of 8/16 thing for jpeg.
    """
    assert angle in (0, 90, 180, 270, -90), "Invalid rotation angle {}".format(angle)
    if angle == 0:
        return
    fnamebase, fnameext = os.path.splitext(fname)
    if fnameext.lower() in (".jpg", ".jpeg"):
        print("**** Doing JPEG rotation {} on {}".format(angle, fname))
        im = jpegtran.JPEGImage(fname)
        im.rotate(angle).save(fname)
        return

    subprocess.run(
        ["mogrify", "-quiet", "-rotate", str(angle), fname],
        stderr=subprocess.STDOUT,
        shell=False,
        check=True,
    )



import peewee as pw

plomdb = pw.SqliteDatabase(None)


class BaseModel(pw.Model):
    class Meta:
        database = plomdb


class User(BaseModel):
    name = pw.CharField(unique=True)
    enabled = pw.BooleanField(default=True)
    password = pw.CharField(null=True)  # hash of password for comparison
    token = pw.CharField(null=True)  # authentication token
    last_activity = pw.DateTimeField(null=True)
    last_action = pw.CharField(null=True)


class Bundle(BaseModel):
    name = pw.CharField(unique=True, null=True)  # unique names please
    md5sum = pw.CharField(null=True)  # to check for duplications


class Image(BaseModel):
    bundle = pw.ForeignKeyField(Bundle, backref="images")
    original_name = pw.CharField(null=True)  # can be empty.
    # the order of the image within its bundle
    bundle_order = pw.IntegerField(null=True)
    file_name = pw.CharField(null=True)
    md5sum = pw.CharField(null=True)  # to check for duplications


class Test(BaseModel):
    test_number = pw.IntegerField(primary_key=True, unique=True)
    # some state pw.Bools
    produced = pw.BooleanField(default=False)
    used = pw.BooleanField(default=False)
    scanned = pw.BooleanField(default=False)
    identified = pw.BooleanField(default=False)
    marked = pw.BooleanField(default=False)
    # a recentUpload flag to see which tests to check after uploads
    recent_upload = pw.BooleanField(default=False)


class Group(BaseModel):
    test = pw.ForeignKeyField(Test, backref="groups")
    gid = pw.CharField(unique=True)  # must be unique
    group_type = pw.CharField()  # to distinguish between ID, DNM, and Mark groups
    queue_position = pw.IntegerField(unique=True, null=False)
    scanned = pw.BooleanField(default=False)  # should get all its tpages
    # a recentUpload flag to see which groups to check after uploads
    recent_upload = pw.BooleanField(default=False)


class IDGroup(BaseModel):
    test = pw.ForeignKeyField(Test, backref="idgroups")
    group = pw.ForeignKeyField(Group, backref="idgroups")
    student_id = pw.CharField(unique=True, null=True)
    student_name = pw.CharField(null=True)
    user = pw.ForeignKeyField(User, backref="idgroups", null=True)
    status = pw.CharField(default="")
    time = pw.DateTimeField(null=True)
    identified = pw.BooleanField(default=False)


class DNMGroup(BaseModel):
    test = pw.ForeignKeyField(Test, backref="dnmgroups")
    group = pw.ForeignKeyField(Group, backref="dnmgroups")


class QGroup(BaseModel):
    test = pw.ForeignKeyField(Test, backref="qgroups")
    group = pw.ForeignKeyField(Group, backref="qgroups")
    question = pw.IntegerField(null=False)
    version = pw.IntegerField(null=False, default=1)
    user = pw.ForeignKeyField(User, backref="qgroups", null=True)
    status = pw.CharField(default="")
    time = pw.DateTimeField(null=True)
    marked = pw.BooleanField(default=False)


class TPage(BaseModel):  # a test page that knows its tpgv
    test = pw.ForeignKeyField(Test, backref="tpages")
    page_number = pw.IntegerField(null=False)
    version = pw.IntegerField(default=1)
    group = pw.ForeignKeyField(Group, backref="tpages")
    image = pw.ForeignKeyField(Image, backref="tpages", null=True)
    scanned = pw.BooleanField(default=False)  # we should get all of them
    # note - Do not delete - rather set scanned=False


class HWPage(BaseModel):  # a hw page that knows its tgv, but not p.
    test = pw.ForeignKeyField(Test, backref="hwpages")
    group = pw.ForeignKeyField(Group, backref="hwpages")
    order = pw.IntegerField(null=False)
    version = pw.IntegerField(default=1)  # infer from group
    image = pw.ForeignKeyField(Image, backref="hwpages")


# an extra page that knows its tgv, but not p. - essentially same as hwpages.
class EXPage(BaseModel):
    test = pw.ForeignKeyField(Test, backref="expages")
    group = pw.ForeignKeyField(Group, backref="expages")
    order = pw.IntegerField(null=False)
    version = pw.IntegerField(default=1)  # infer from group
    image = pw.ForeignKeyField(Image, backref="expages")


class LPage(BaseModel):  # a page that just knows its t. - a loose page
    test = pw.ForeignKeyField(Test, backref="lpages")
    order = pw.IntegerField(null=False)
    image = pw.ForeignKeyField(Image, backref="lpages")


# still needs work - maybe some bundle object with unique key.
class UnknownPage(BaseModel):
    image = pw.ForeignKeyField(Image, backref="upages", null=True)
    order = pw.IntegerField(null=False)  # order within the upload.


class CollidingPage(BaseModel):
    tpage = pw.ForeignKeyField(TPage, backref="collisions")
    image = pw.ForeignKeyField(Image, backref="collisions")


class DiscardedPage(BaseModel):
    image = pw.ForeignKeyField(Image, backref="discards")
    reason = pw.CharField(null=True)


class IDPage(BaseModel):
    idgroup = pw.ForeignKeyField(IDGroup, backref="idpages")
    image = pw.ForeignKeyField(Image, backref="idpages")
    order = pw.IntegerField(null=False)


class DNMPage(BaseModel):
    dnmgroup = pw.ForeignKeyField(DNMGroup, backref="dnmpages")
    image = pw.ForeignKeyField(Image, backref="dnmpages")
    order = pw.IntegerField(null=False)


class AImage(BaseModel):  # a class for containing annotation-images
    file_name = pw.CharField(null=True)
    md5sum = pw.CharField(null=True)  # to check for duplications


class Annotation(BaseModel):
    qgroup = pw.ForeignKeyField(QGroup, backref="annotations")
    user = pw.ForeignKeyField(User, backref="annotations", null=True)
    aimage = pw.ForeignKeyField(AImage, backref="annotations", null=True)
    edition = pw.IntegerField(null=True)
    integrity_check = pw.CharField(null=True)  # random uuid
    # we need to order the annotations - want the latest.
    plom_file = pw.CharField(null=True)
    comment_file = pw.CharField(null=True)
    mark = pw.IntegerField(null=True)
    marking_time = pw.IntegerField(null=True)
    time = pw.DateTimeField(null=True)
    tags = pw.CharField(default="")


class APage(BaseModel):
    annotation = pw.ForeignKeyField(Annotation, backref="apages")
    image = pw.ForeignKeyField(Image, backref="apages")
    order = pw.IntegerField(null=False)


class OldAnnotation(BaseModel):
    qgroup = pw.ForeignKeyField(QGroup, backref="oldannotations")
    user = pw.ForeignKeyField(User, backref="oldannotations", null=True)
    aimage = pw.ForeignKeyField(AImage, backref="oldannotations", null=True)
    edition = pw.IntegerField(null=True)
    integrity_check = pw.CharField(null=True)  # concat of md5sums of underlying apages
    # we need to order the annotations - want the latest.
    plom_file = pw.CharField(null=True)
    comment_file = pw.CharField(null=True)
    mark = pw.IntegerField(null=True)
    marking_time = pw.IntegerField(null=True)
    time = pw.DateTimeField(null=True)
    tags = pw.CharField(default="")


class OAPage(BaseModel):
    old_annotation = pw.ForeignKeyField(OldAnnotation, backref="oapages")
    image = pw.ForeignKeyField(Image, backref="oapages")
    order = pw.IntegerField(null=False)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

import os
import sys
from getpass import getpass

from plom import SpecParser
from plom import specdir
from plom.produce import build_all_papers, confirm_processed, identify_prenamed
from plom.produce import paperdir
from plom.messenger import ManagerMessenger
from plom.plom_exceptions import PlomExistingLoginException, PlomBenignException


def buildDatabaseAndPapers(server=None, password=None, fakepdf=False, no_qr=False):
    if server and ":" in server:
        s, p = server.split(":")
        msgr = ManagerMessenger(s, port=p)
    else:
        msgr = ManagerMessenger(server)
    msgr.start()

    if not password:
        password = getpass('Please enter the "manager" password: ')

    try:
        msgr.requestAndSaveToken("manager", password)
    except PlomExistingLoginException:
        # TODO: bit annoying, maybe want manager UI open...
        print(
            "You appear to be already logged in!\n\n"
            "  * Perhaps a previous session crashed?\n"
            "  * Do you have another management tool running,\n"
            "    e.g., on another computer?\n\n"
            'In order to force-logout the existing authorisation run "plom-build clear"'
        )
        exit(10)
    try:
        spec = msgr.get_spec()
        try:
            status = msgr.TriggerPopulateDB()
        except PlomBenignException:
            print("Error: Server already has a populated database")
            exit(3)
        print(status)
        pvmap = msgr.getGlobalPageVersionMap()
        os.makedirs(paperdir, exist_ok=True)

        if spec["numberToName"] > 0:
            try:
                classlist = msgr.IDrequestClasslist()
            except PlomBenignException as e:
                print("Failed to download classlist: {}".format(e))
                exit(4)
            print(
                'Building {} pre-named papers and {} blank papers in "{}"...'.format(
                    spec["numberToName"],
                    spec["numberToProduce"] - spec["numberToName"],
                    paperdir,
                )
            )
        else:
            classlist = None
            print(
                'Building {} blank papers in "{}"...'.format(
                    spec["numberToProduce"], paperdir
                )
            )
        build_all_papers(spec, pvmap, classlist, fakepdf, no_qr=no_qr)

        print("Checking papers produced and updating databases")
        confirm_processed(spec, msgr, classlist)
        print("Identifying any pre-named papers into the database")
        identify_prenamed(spec, msgr, classlist)
    finally:
        msgr.closeUser()
        msgr.stop()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './qtCreatorFiles/ui_identify.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_IdentifyWindow(object):
    def setupUi(self, IdentifyWindow):
        IdentifyWindow.setObjectName("IdentifyWindow")
        IdentifyWindow.setWindowModality(QtCore.Qt.ApplicationModal)
        IdentifyWindow.resize(1000, 600)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(IdentifyWindow.sizePolicy().hasHeightForWidth())
        IdentifyWindow.setSizePolicy(sizePolicy)
        IdentifyWindow.setBaseSize(QtCore.QSize(0, 0))
        self.gridLayout_5 = QtWidgets.QGridLayout(IdentifyWindow)
        self.gridLayout_5.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.closeButton = QtWidgets.QPushButton(IdentifyWindow)
        self.closeButton.setObjectName("closeButton")
        self.horizontalLayout.addWidget(self.closeButton)
        self.gridLayout_5.addLayout(self.horizontalLayout, 4, 0, 1, 1)
        self.widget_2 = QtWidgets.QWidget(IdentifyWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(5)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget_2.sizePolicy().hasHeightForWidth())
        self.widget_2.setSizePolicy(sizePolicy)
        self.widget_2.setObjectName("widget_2")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.widget_2)
        self.gridLayout_4.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.paperBox = QtWidgets.QGroupBox(self.widget_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(4)
        sizePolicy.setHeightForWidth(self.paperBox.sizePolicy().hasHeightForWidth())
        self.paperBox.setSizePolicy(sizePolicy)
        self.paperBox.setObjectName("paperBox")
        self.gridLayout_7 = QtWidgets.QGridLayout(self.paperBox)
        self.gridLayout_7.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.gridLayout_4.addWidget(self.paperBox, 1, 0, 1, 1)
        self.predictionBox = QtWidgets.QGroupBox(self.widget_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.predictionBox.sizePolicy().hasHeightForWidth())
        self.predictionBox.setSizePolicy(sizePolicy)
        self.predictionBox.setObjectName("predictionBox")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.predictionBox)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.pSIDLabel = QtWidgets.QLabel(self.predictionBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pSIDLabel.sizePolicy().hasHeightForWidth())
        self.pSIDLabel.setSizePolicy(sizePolicy)
        self.pSIDLabel.setObjectName("pSIDLabel")
        self.gridLayout_6.addWidget(self.pSIDLabel, 0, 0, 1, 1)
        self.pNameLabel = QtWidgets.QLabel(self.predictionBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(4)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pNameLabel.sizePolicy().hasHeightForWidth())
        self.pNameLabel.setSizePolicy(sizePolicy)
        self.pNameLabel.setObjectName("pNameLabel")
        self.gridLayout_6.addWidget(self.pNameLabel, 1, 0, 1, 1)
        self.predButton = QtWidgets.QPushButton(self.predictionBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(1)
        sizePolicy.setHeightForWidth(self.predButton.sizePolicy().hasHeightForWidth())
        self.predButton.setSizePolicy(sizePolicy)
        self.predButton.setObjectName("predButton")
        self.gridLayout_6.addWidget(self.predButton, 0, 1, 2, 1)
        self.gridLayout_4.addWidget(self.predictionBox, 0, 0, 1, 1)
        self.gridLayout_5.addWidget(self.widget_2, 0, 1, 5, 1)
        self.widget = QtWidgets.QWidget(IdentifyWindow)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy)
        self.widget.setObjectName("widget")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.widget)
        self.gridLayout_3.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.nextButton = QtWidgets.QPushButton(self.widget)
        self.nextButton.setObjectName("nextButton")
        self.gridLayout_3.addWidget(self.nextButton, 3, 0, 1, 1)
        self.userBox = QtWidgets.QGroupBox(self.widget)
        self.userBox.setObjectName("userBox")
        self.gridLayout = QtWidgets.QGridLayout(self.userBox)
        self.gridLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout.setObjectName("gridLayout")
        self.userLabel = QtWidgets.QLabel(self.userBox)
        self.userLabel.setObjectName("userLabel")
        self.gridLayout.addWidget(self.userLabel, 0, 0, 1, 1)
        self.gridLayout_3.addWidget(self.userBox, 0, 0, 1, 1)
        self.tableBox = QtWidgets.QGroupBox(self.widget)
        self.tableBox.setObjectName("tableBox")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.tableBox)
        self.gridLayout_2.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.tableView = QtWidgets.QTableView(self.tableBox)
        self.tableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.tableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView.setObjectName("tableView")
        self.tableView.horizontalHeader().setStretchLastSection(True)
        self.gridLayout_2.addWidget(self.tableView, 0, 0, 1, 1)
        self.gridLayout_3.addWidget(self.tableBox, 5, 0, 1, 1)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.blankButton = QtWidgets.QPushButton(self.widget)
        self.blankButton.setObjectName("blankButton")
        self.horizontalLayout_2.addWidget(self.blankButton)
        self.viewButton = QtWidgets.QPushButton(self.widget)
        self.viewButton.setObjectName("viewButton")
        self.horizontalLayout_2.addWidget(self.viewButton)
        self.gridLayout_3.addLayout(self.horizontalLayout_2, 4, 0, 1, 1)
        self.studentBox = QtWidgets.QGroupBox(self.widget)
        self.studentBox.setObjectName("studentBox")
        self.formLayout = QtWidgets.QFormLayout(self.studentBox)
        self.formLayout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.formLayout.setObjectName("formLayout")
        self.idEdit = QtWidgets.QLineEdit(self.studentBox)
        self.idEdit.setObjectName("idEdit")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.idEdit)
        self.gridLayout_3.addWidget(self.studentBox, 2, 0, 1, 1)
        self.gridLayout_5.addWidget(self.widget, 0, 0, 1, 1)
        self.progressGroupBox = QtWidgets.QGroupBox(IdentifyWindow)
        self.progressGroupBox.setObjectName("progressGroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.progressGroupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.idProgressBar = QtWidgets.QProgressBar(self.progressGroupBox)
        self.idProgressBar.setMaximum(1)
        self.idProgressBar.setProperty("value", 0)
        self.idProgressBar.setObjectName("idProgressBar")
        self.verticalLayout.addWidget(self.idProgressBar)
        self.gridLayout_5.addWidget(self.progressGroupBox, 1, 0, 1, 1)

        self.retranslateUi(IdentifyWindow)
        QtCore.QMetaObject.connectSlotsByName(IdentifyWindow)
        IdentifyWindow.setTabOrder(self.nextButton, self.tableView)
        IdentifyWindow.setTabOrder(self.tableView, self.closeButton)

    def retranslateUi(self, IdentifyWindow):
        _translate = QtCore.QCoreApplication.translate
        IdentifyWindow.setWindowTitle(_translate("IdentifyWindow", "Identify papers"))
        self.closeButton.setText(_translate("IdentifyWindow", "&Close"))
        self.paperBox.setTitle(_translate("IdentifyWindow", "Current paper"))
        self.predictionBox.setTitle(_translate("IdentifyWindow", "Machine prediction"))
        self.pSIDLabel.setText(_translate("IdentifyWindow", "Predicted Student ID"))
        self.pNameLabel.setText(_translate("IdentifyWindow", "Predicted Student Name"))
        self.predButton.setText(_translate("IdentifyWindow", "Accept\n"
" machine\n"
" prediction"))
        self.nextButton.setText(_translate("IdentifyWindow", "Skip (for now) and &get next"))
        self.userBox.setTitle(_translate("IdentifyWindow", "User"))
        self.userLabel.setText(_translate("IdentifyWindow", "Username"))
        self.tableBox.setTitle(_translate("IdentifyWindow", "Table of papers"))
        self.blankButton.setText(_translate("IdentifyWindow", "&Blank page"))
        self.viewButton.setText(_translate("IdentifyWindow", "&View whole"))
        self.studentBox.setTitle(_translate("IdentifyWindow", "Search classlist by student ID or name"))
        self.progressGroupBox.setTitle(_translate("IdentifyWindow", "Progress"))
        self.idProgressBar.setFormat(_translate("IdentifyWindow", "%v of %m"))



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2019 Andrew Rechnitzer
# Copyright (C) 2020 Victoria Schuster

__copyright__ = "Copyright (C) 2018-2019 Andrew Rechnitzer and others"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Victoria Schuster"]
__license__ = "AGPLv3"

import logging

log = logging.getLogger("tools")

from plom.client.tools.arrow import *
from plom.client.tools.box import *
from plom.client.tools.comment import *
from plom.client.tools.cross import *
from plom.client.tools.delete import *
from plom.client.tools.delta import *
from plom.client.tools.ellipse import *
from plom.client.tools.highlight import *
from plom.client.tools.image import *
from plom.client.tools.line import *
from plom.client.tools.move import *
from plom.client.tools.pen import *
from plom.client.tools.penArrow import *
from plom.client.tools.questionMark import *
from plom.client.tools.text import *
from plom.client.tools.tick import *



#!/usr/bin/env python3
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald

"""Randomly ID papers for testing purposes."""

__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer and others"
__credits__ = "The Plom Project Developers"
__license__ = "AGPL-3.0-or-later"

import argparse
import getpass
import random

from plom.plom_exceptions import (
    PlomConflict,
    PlomExistingLoginException,
    PlomTakenException,
)
from plom.messenger import Messenger

# -------------------------------------------


def startIdentifying():
    idList = messenger.IDrequestClasslist()

    while True:
        task = messenger.IDaskNextTask()
        if task is None:
            print("No more tasks.")
            break
        try:
            print("Identifying task ", task)
            imageList = messenger.IDclaimThisTask(task)
        except PlomTakenException:
            # task already taken.
            continue

        while True:
            try:
                sid, sname = random.choice(idList)
                sname += " [randomly chosen]"
                messenger.IDreturnIDdTask(task, sid, sname)
                break
            except PlomConflict:
                print("SID/SN {}/{} already used".format(sid, sname))


# -------------------------------------------

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Perform identifier tasks randomly, generally for testing."
    )

    parser.add_argument("-w", "--password", type=str)
    parser.add_argument("-u", "--user", type=str)
    parser.add_argument(
        "-s",
        "--server",
        metavar="SERVER[:PORT]",
        action="store",
        help="Which server to contact.",
    )
    global messenger
    args = parser.parse_args()
    if args.server and ":" in args.server:
        s, p = args.server.split(":")
        messenger = Messenger(s, port=p)
    else:
        messenger = Messenger(args.server)
    messenger.start()

    # If user not specified then default to scanner
    if args.user is None:
        user = "scanner"
    else:
        user = args.user

    # get the password if not specified
    if args.password is None:
        pwd = getpass.getpass("Please enter the '{}' password:".format(user))
    else:
        pwd = args.password

    # get started
    try:
        messenger.requestAndSaveToken(user, pwd)
    except PlomExistingLoginException:
        print(
            "You appear to be already logged in!\n\n"
            "  * Perhaps a previous session crashed?\n"
            "  * Do you have another scanner-script running,\n"
            "    e.g., on another computer?\n\n"
            "This script has automatically force-logout'd that user."
        )
        messenger.clearAuthorisation(user, pwd)
        exit(1)

    spec = messenger.get_spec()

    try:
        startIdentifying()
    except Exception as e:
        print("Error identifying papers: {}".format(e))
        exit(1)

    messenger.closeUser()
    messenger.stop()



# -*- coding: utf-8 -*-

__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2019-2020 Andrew Rechnitzer and Colin Macdonald"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald"]
__license__ = "AGPL-3.0-or-later"
# SPDX-License-Identifier: AGPL-3.0-or-later

import getpass

from plom.messenger import ScanMessenger


def clearLogin(server=None, password=None):
    if server and ":" in server:
        s, p = server.split(":")
        scanMessenger = ScanMessenger(s, port=p)
    else:
        scanMessenger = ScanMessenger(server)
    scanMessenger.start()

    # get the password if not specified
    if password is None:
        try:
            pwd = getpass.getpass("Please enter the 'scanner' password:")
        except Exception as error:
            print("ERROR", error)
            exit(1)
    else:
        pwd = password

    scanMessenger.clearAuthorisation("scanner", pwd)
    print("Scanner login cleared.")
    scanMessenger.stop()



__author__ = "Andrew Rechnitzer"
__copyright__ = "Copyright (C) 2020 Andrew Rechnitzer"
__credits__ = ["Andrew Rechnitzer", "Colin Macdonald", "Elvis Cai", "Matt Coles"]
__license__ = "AGPLv3"

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush, QGuiApplication, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QDialog,
    QFrame,
    QFormLayout,
    QGraphicsPixmapItem,
    QGraphicsItemGroup,
    QGraphicsScene,
    QGraphicsView,
    QGridLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)


class ActionTab(QWidget):
    def __init__(self, parent):
        super(ActionTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.rb = QPushButton("Review")
        self.nb = QPushButton("No action")
        vb.addWidget(self.rb)
        vb.addStretch(0)
        vb.addWidget(self.nb)
        vb.addStretch(0)
        self.setLayout(vb)
        self.show()
        self.rb.clicked.connect(self.review)
        self.nb.clicked.connect(self.noaction)

    def review(self):
        self.parent.optionTW.setCurrentIndex(1)

    def noaction(self):
        self.parent.action = "none"
        self.parent.accept()


class ReviewTab(QWidget):
    def __init__(self, parent):
        super(ReviewTab, self).__init__()
        self.parent = parent
        vb = QVBoxLayout()
        self.rb = QPushButton("Click to confirm review")
        self.ob = QPushButton("Return to other options")
        vb.addStretch(0)
        vb.addWidget(self.rb)
        vb.addStretch(0)
        vb.addWidget(self.ob)
        self.setLayout(vb)
        self.show()
        self.rb.clicked.connect(self.review)
        self.ob.clicked.connect(self.other)

    def review(self):
        self.parent.action = "review"
        self.parent.accept()

    def other(self):
        self.parent.action = "none"
        self.parent.optionTW.setCurrentIndex(0)


class ReviewViewWindow(QDialog):
    """Simple view window for pageimages"""

    def __init__(self, parent, fnames, quidto="question"):
        QWidget.__init__(self)
        self.parent = parent
        self.quidto = quidto

        if type(fnames) == list:
            self.initUI(fnames)
        else:
            self.initUI([fnames])
        self.action = "none"

    def initUI(self, fnames):
        self.view = ReviewView(fnames)
        # Render nicely
        self.view.setRenderHint(QPainter.HighQualityAntialiasing)
        self.optionTW = QTabWidget()

        # reset view button passes to the UnknownView.
        self.resetB = QPushButton("reset view")
        self.cancelB = QPushButton("&cancel")
        self.maxNormB = QPushButton("&max/norm")

        self.cancelB.clicked.connect(self.reject)
        self.resetB.clicked.connect(lambda: self.view.resetView())
        self.maxNormB.clicked.connect(self.swapMaxNorm)

        self.resetB.setAutoDefault(False)  # return wont click the button by default.

        # Layout simply
        grid = QGridLayout()
        grid.addWidget(self.view, 1, 1, 10, 6)
        grid.addWidget(self.optionTW, 2, 17, 8, 4)
        grid.addWidget(self.resetB, 20, 1)
        grid.addWidget(self.cancelB, 20, 20)
        grid.addWidget(self.maxNormB, 1, 20)
        self.setLayout(grid)
        self.show()
        # Store the current exam view as a qtransform
        self.viewTrans = self.view.transform()
        self.dx = self.view.horizontalScrollBar().value()
        self.dy = self.view.verticalScrollBar().value()
        self.theta = 0
        self.initTabs()

    def updateImage(self, fnames):
        """Pass file to the view to update the image"""
        # first store the current view transform and scroll values
        self.viewTrans = self.view.transform()
        self.dx = self.view.horizontalScrollBar().value()
        self.dy = self.view.verticalScrollBar().value()
        # update the image
        if type(fnames) == list:
            self.view.updateImage(fnames)
        else:
            self.view.updateImage([fnames])

        # re-set the view transform and scroll values
        self.view.setTransform(self.viewTrans)
        self.view.horizontalScrollBar().setValue(self.dx)
        self.view.verticalScrollBar().setValue(self.dy)

    def initTabs(self):
        self.t0 = ActionTab(self)
        self.t1 = ReviewTab(self)
        self.optionTW.addTab(self.t0, "Actions")
        self.optionTW.addTab(self.t1, "Review {}".format(self.quidto))

    def swapMaxNorm(self):
        """Toggles the window size between max and normal"""
        if self.windowState() != Qt.WindowMaximized:
            self.setWindowState(Qt.WindowMaximized)
        else:
            self.setWindowState(Qt.WindowNoState)


class ReviewView(QGraphicsView):
    """Simple extension of QGraphicsView
    - containing an image and click-to-zoom/unzoom
    """

    def __init__(self, fnames):
        QGraphicsView.__init__(self)
        self.initUI(fnames)

    def initUI(self, fnames):
        # Make QGraphicsScene
        self.scene = QGraphicsScene()
        # TODO = handle different image sizes.
        self.images = {}
        self.imageGItem = QGraphicsItemGroup()
        self.scene.addItem(self.imageGItem)
        self.updateImage(fnames)
        self.setBackgroundBrush(QBrush(Qt.darkCyan))

    def updateImage(self, fnames):
        """Update the image with that from filename"""
        for n in self.images:
            self.imageGItem.removeFromGroup(self.images[n])
            self.images[n].setVisible(False)
        if fnames is not None:
            x = 0
            n = 0
            for fn in fnames:
                self.images[n] = QGraphicsPixmapItem(QPixmap(fn))
                self.images[n].setTransformationMode(Qt.SmoothTransformation)
                self.images[n].setPos(x, 0)
                self.images[n].setVisible(True)
                self.scene.addItem(self.images[n])
                x += self.images[n].boundingRect().width() + 10
                self.imageGItem.addToGroup(self.images[n])
                n += 1

        # Set sensible sizes and put into the view, and fit view to the image.
        br = self.imageGItem.boundingRect()
        self.scene.setSceneRect(
            0,
            0,
            max(1000, br.width()),
            max(1000, br.height()),
        )
        self.setScene(self.scene)
        self.fitInView(self.imageGItem, Qt.KeepAspectRatio)

    def mouseReleaseEvent(self, event):
        """Left/right click to zoom in and out"""
        if (event.button() == Qt.RightButton) or (
            QGuiApplication.queryKeyboardModifiers() == Qt.ShiftModifier
        ):
            self.scale(0.8, 0.8)
        else:
            self.scale(1.25, 1.25)
        self.centerOn(event.pos())

    def resetView(self):
        """Reset the view to its reasonable initial state."""
        self.fitInView(self.imageGItem, Qt.KeepAspectRatio)



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Dryden Wiebe
# Copyright (C) 2020 Vala Vakilian
# Copyright (C) 2020 Colin B. Macdonald

"""
Note: Code in this file is very similar to predictStudentID code for the Sklearn
model.

See:
https://www.pyimagesearch.com/2017/02/13/recognizing-digits-with-opencv-and-python/
"""

import cv2
import imutils
from imutils.perspective import four_point_transform
import numpy as np
import tensorflow as tf

# hack suggested here https://github.com/tensorflow/tensorflow/issues/34201
# import tensorflow.keras.backend as K
import tensorflow.python.keras.backend as K


# from https://fairyonice.github.io/Measure-the-uncertainty-in-deep-learning-models-using-dropout.html
class keras_dropout_prediction_object(object):
    def __init__(self, model):
        # setup a function to predict the digit with dropout (default is to predict without dropout)
        self.f = K.function(
            [model.layers[0].input, K.symbolic_learning_phase()],
            [model.output],
        )

    def predict_with_dropout(self, x, n_iter=20):
        """Calculates digit predictions with dropout.

        Arguments:
            x (np.array): The image that we are intersted in.

        Keyword Arguments:
            n_iter (int): Number of iterations to run the predictions with dropout.
                Defaults to 20.

        Returns:
            np.array: The digit predictions for each forward pass of the model,
                dimensions [10 (for the 10 digit classes), n_iter].
        """
        result = []
        for _ in range(n_iter):
            result.append(self.f([x, True]))
        result = np.array(result).reshape(n_iter, 10).T
        return result


# define this in order to sort by area of bounding rect
def bounding_rect_area(bounding_rectangle):
    """Return the area of the rectangle.

    Args:
        bounding_rectangle (list): Target rectangle object.

    Returns:
        int: Area of the rectangle.
    """
    _, _, w, h = cv2.boundingRect(bounding_rectangle)
    return w * h


def get_digit_box(image_box_fname, top, bottom):
    """Find the box that includes the student ID for extracting the digits.

    Args:
        image_box_fname (str): Name of the front image.
        top (int): Top coordinate of the cropping.
        bottom (int): Bottom coordinate of the cropping.

    Returns:
        numpy.ndarray: The processed image that includes the student
            ID digits.
    """

    # Read in the whole cropped image
    front_image = cv2.imread(image_box_fname)

    # TODO: Why do we even pass top and bottom.
    # Extract only the required portion of the image.
    cropped_image = front_image[:][top:bottom]

    # Process the image so as to find the countours.
    # Grey, Blur and Edging are standard processes for text detection.
    # TODO: I must make these better formatted.
    grey_image = cv2.cvtColor(cropped_image, cv2.COLOR_BGR2GRAY)
    blurred_image = cv2.GaussianBlur(grey_image, (5, 5), 0)
    edged_image = cv2.Canny(blurred_image, 50, 200, 255)

    # Find the contours to find the black bordered box.
    contours = cv2.findContours(
        edged_image.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
    )
    contour_lists = imutils.grab_contours(contours)
    sorted_contour_list = sorted(contour_lists, key=cv2.contourArea, reverse=True)
    id_box_contour = None

    # Loop over the contours.
    for contour in sorted_contour_list:
        # Approximate the contour.
        perimeter = cv2.arcLength(contour, True)
        third_order_moment = cv2.approxPolyDP(contour, 0.02 * perimeter, True)

        if len(third_order_moment) == 4:
            id_box_contour = third_order_moment
            break

    # TODO: Why is this this not using edged_image
    # warped = four_point_transform(edged_image, id_box_contour.reshape(4, 2))
    output = four_point_transform(cropped_image, id_box_contour.reshape(4, 2))

    # TODO: Remove magic number creation.
    # note that this width of 1250 is defined by the IDbox template
    new_width = int(output.shape[0] * 1250.0 / output.shape[1])
    scaled = cv2.resize(output, (1250, new_width), cv2.INTER_CUBIC)

    # the digit box numbers again come from the IDBox template and numerology
    ID_box = scaled[30:300, 355:1220]

    # TODO: This is for debugging
    # cv2.imwrite('goddamns' + '.png', ID_box)
    # input("Press Enter...")

    return ID_box


def get_digit_images(ID_box, num_digits):
    """Find the digit images and return them in a list.

    Args:
        ID_box (class 'numpy.ndarray): Image containing the student ID.
        num_digits (int): Number of digits in the student ID.

    Returns:
        list: A list of numpy.ndarray which are the images for each digit.
    """

    processed_digits_images_list = []

    for digit_index in range(num_digits):

        # TODO: Maybe remove magical hackery.
        # extract the kth digit box. Some magical hackery / numerology here.
        digit1 = ID_box[0:250, digit_index * 109 + 5 : (digit_index + 1) * 109 - 5]

        # TODO: I think I could remove all of this.
        # Now some hackery to centre on the digit so closer to mnist dataset.
        # Find the contours and centre on the largest (by area).
        digit2 = cv2.GaussianBlur(digit1, (3, 3), 0)
        digit3 = cv2.Canny(digit2, 5, 255, 200)
        contours = cv2.findContours(
            digit3.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        contour_lists = imutils.grab_contours(contours)
        # sort by bounding box area
        sorted_contours = sorted(contour_lists, key=bounding_rect_area, reverse=True)
        # make sure we can find at least one contour
        if len(sorted_contours) == 0:
            # can't make a prediction so return
            return None
        # get bounding rect of biggest contour
        bnd = cv2.boundingRect(sorted_contours[0])
        # put some padding around that rectangle
        pad = 10
        xl = max(0, bnd[1] - pad)
        yt = max(0, bnd[0] - pad)
        # grab the image - should be the digit.
        digit4 = digit2[xl : bnd[1] + bnd[3] + pad, yt : bnd[0] + bnd[2] + pad]
        # Do some clean-up by thresholding pixels
        digit5 = cv2.adaptiveThreshold(
            cv2.cvtColor(digit4, cv2.COLOR_BGR2GRAY),
            255,
            cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
            cv2.THRESH_BINARY_INV,
            31,
            1,
        )
        # and a little more - blur helps get rid of "dust" artifacts
        digit6 = cv2.blur(digit5, (3, 3))
        # now need to resize it to height or width =28 (depending on aspect ratio)
        # the "28" comes from mnist dataset, mnist digits are 28 x 28
        rat = digit5.shape[0] / digit5.shape[1]
        if rat > 1:
            w = 28
            h = int(28 // rat)
        else:
            w = int(28 * rat)
            h = 28
        # region of interest
        roi = cv2.resize(digit6, (h, w), interpolation=cv2.INTER_AREA)
        px = int((28 - w) // 2)
        py = int((28 - h) // 2)
        # and a bit more clean-up - put black around border where needed
        roi2 = cv2.copyMakeBorder(
            roi, px, 28 - w - px, py, 28 - h - py, cv2.BORDER_CONSTANT, value=[0, 0, 0]
        )

        # get it into format needed by model predictor
        roi3 = np.expand_dims(roi2, 0)
        # do the actual prediction! (ie approx probabilities that image is digit 0,1,2,..,9)

        roi3 = roi3.reshape((1, 28 * 28))

        processed_digits_images_list.append(roi3)

    return processed_digits_images_list


def get_digit_prob(prediction_model, image_box_fname, top, bottom, num_digits):
    """Return a list of probability predictions for the student ID digits on the cropped image.

    Args:
        prediction_model (keras_dropout_prediction_object): Prediction model.
        image_box_fname (Str): File path for the cropped image which includes the ID box.
        top (int): Top boundary of image.
        bottom (int): Bottom boundary of image.
        num_digits (int): Number of digits in the student ID.

    Returns:
        list: A list of lists of probabilities including the model's
            prediction for the digits.
    """

    # Retrieve the box including the digits in a row.
    ID_box = get_digit_box(image_box_fname, top, bottom)

    # The list of the resulting probabilities.
    prob_lists = []

    processed_digits_images_list = get_digit_images(ID_box, num_digits)

    # and append that prediction to list
    for processed_digit_image in processed_digits_images_list:
        number_pred_prob = prediction_model.predict_with_dropout(
            processed_digit_image
        ).mean(axis=1)
        prob_lists.append(number_pred_prob)

    return prob_lists


def compute_probabilities(
    image_file_paths, top_coordinate, bottom_coordinate, num_digits
):
    """Return a list of probabilities for digits for each test.

    Args:
        image_file_paths (dict): A dictionary including the paths of the images.
        top_coordinate (int): Top boundary of image.
        bottom_coordinate (int): Bottom boundary of image.
        num_digits (int): Number of digits in the student ID.

    Returns:
        dict: A dictionary which involves the probabilities for each image file.
    """

    # load the model
    # fire up the model
    model = tf.keras.models.load_model("model_cache")
    model.compile(
        optimizer="adam",
        loss="sparse_categorical_crossentropy",  # we are using integer classes
        metrics=[tf.keras.metrics.SparseCategoricalAccuracy()],
    )
    prediction_model = keras_dropout_prediction_object(model)

    # Dictionary of test numbers their digit-probabilities
    probabilities = {}

    for testNumber in image_file_paths:
        prob_lists = get_digit_prob(
            prediction_model,
            image_file_paths[testNumber],
            top_coordinate,
            bottom_coordinate,
            num_digits,
        )
        if prob_lists is None:  # couldn't recognize digits
            continue
        probabilities[testNumber] = prob_lists

    return probabilities



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2018-2020 Andrew Rechnitzer
# Copyright (C) 2020 Colin B. Macdonald
# Copyright (C) 2020 Victoria Schuster

from math import sqrt

from PyQt5.QtCore import QTimer, QPropertyAnimation, pyqtProperty, Qt, QPointF
from PyQt5.QtGui import QPen, QPainterPath, QBrush, QColor
from PyQt5.QtWidgets import (
    QUndoCommand,
    QGraphicsObject,
    QGraphicsItemGroup,
    QGraphicsPathItem,
    QGraphicsItem,
)

from plom.client.tools import CommandMoveItem, log


class CommandPenArrow(QUndoCommand):
    # Very similar to CommandArrow
    def __init__(self, scene, path):
        super(CommandPenArrow, self).__init__()
        self.scene = scene
        self.path = path
        self.penItem = PenArrowItemObject(self.path)
        self.setText("PenArrow")

    @classmethod
    def from_pickle(cls, X, *, scene):
        """Construct a CommandPenArrow from a serialized form.

        Raises:
            ValueError: malformed or otherwise incorrect data
            AssertionError: there is a bug somewhere.

        TODO: all these Pen-like annotations should subclass pen
        and inherit this function.
        """
        assert X[0] == "PenArrow"
        X = X[1:]
        if len(X) != 1:
            raise ValueError("wrong length of pickle data")
        # Format is X = [['m',x,y], ['l',x,y], ['l',x,y], ...]
        X = X[0]
        pth = QPainterPath()
        # unpack ['m', x, y] or ValueError
        cmd, x, y = X[0]
        if cmd != "m":
            raise ValueError("malformed start of Pen-like annotation")
        pth.moveTo(QPointF(x, y))
        for pt in X[1:]:
            # unpack ['l', x, y] or ValueError
            cmd, x, y = pt
            if cmd != "l":
                raise ValueError("malformed Pen-like annotation in interior")
            pth.lineTo(QPointF(x, y))
        return cls(scene, pth)

    def redo(self):
        self.penItem.flash_redo()
        self.scene.addItem(self.penItem.pi)

    def undo(self):
        self.penItem.flash_undo()
        QTimer.singleShot(200, lambda: self.scene.removeItem(self.penItem.pi))


class PenArrowItemObject(QGraphicsObject):
    # As per the ArrowItemObject
    def __init__(self, path):
        super(PenArrowItemObject, self).__init__()
        self.pi = PenArrowItem(path, self)
        self.anim = QPropertyAnimation(self, b"thickness")

    def flash_undo(self):
        self.anim.setDuration(200)
        self.anim.setStartValue(2)
        self.anim.setKeyValueAt(0.5, 6)
        self.anim.setEndValue(0)
        self.anim.start()

    def flash_redo(self):
        self.anim.setDuration(200)
        self.anim.setStartValue(2)
        self.anim.setKeyValueAt(0.5, 4)
        self.anim.setEndValue(2)
        self.anim.start()

    @pyqtProperty(int)
    def thickness(self):
        return self.pi.pi.pen().width()

    @thickness.setter
    def thickness(self, value):
        self.pi.pi.setPen(QPen(Qt.red, value))
        self.pi.endi.setPen(QPen(Qt.red, value))
        self.pi.endf.setPen(QPen(Qt.red, value))


class PenArrowItem(QGraphicsItemGroup):
    def __init__(self, path, parent=None):
        super(PenArrowItem, self).__init__()
        self.saveable = True
        self.pi = QGraphicsPathItem()
        self.path = path
        self.animator = [parent]
        self.animateFlag = False

        # set arrowhead initial
        e0 = self.path.elementAt(0)
        e1 = self.path.elementAt(1)
        pti = QPointF(e1.x, e1.y)
        ptf = QPointF(e0.x, e0.y)
        delta = ptf - pti
        el = sqrt(delta.x() ** 2 + delta.y() ** 2)
        ndelta = delta / el
        northog = QPointF(-ndelta.y(), ndelta.x())
        arBase = ptf - 16 * ndelta
        arTip = ptf + 8 * ndelta
        arLeft = arBase - 10 * northog - 4 * ndelta
        arRight = arBase + 10 * northog - 4 * ndelta
        self.ari = QPainterPath()
        self.ari.moveTo(ptf)
        self.ari.lineTo(arLeft)
        self.ari.lineTo(arBase)
        self.ari.lineTo(arRight)
        self.ari.lineTo(ptf)
        self.endi = QGraphicsPathItem()
        self.endi.setPath(self.ari)
        # set arrowhead final
        e2 = self.path.elementAt(self.path.elementCount() - 2)
        e3 = self.path.elementAt(self.path.elementCount() - 1)
        pti = QPointF(e2.x, e2.y)
        ptf = QPointF(e3.x, e3.y)
        delta = ptf - pti
        el = sqrt(delta.x() ** 2 + delta.y() ** 2)
        ndelta = delta / el
        northog = QPointF(-ndelta.y(), ndelta.x())
        arBase = ptf - 16 * ndelta
        arTip = ptf + 8 * ndelta
        arLeft = arBase - 10 * northog - 4 * ndelta
        arRight = arBase + 10 * northog - 4 * ndelta
        self.arf = QPainterPath()
        self.arf.moveTo(ptf)
        self.arf.lineTo(arLeft)
        self.arf.lineTo(arBase)
        self.arf.lineTo(arRight)
        self.arf.lineTo(ptf)
        self.endf = QGraphicsPathItem()
        self.endf.setPath(self.arf)
        # put everything together
        self.pi.setPath(self.path)
        self.pi.setPen(QPen(Qt.red, 2))
        self.endi.setPen(QPen(Qt.red, 2))
        self.endi.setBrush(QBrush(Qt.red))
        self.endf.setPen(QPen(Qt.red, 2))
        self.endf.setBrush(QBrush(Qt.red))
        self.setFlag(QGraphicsItem.ItemIsMovable)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges)
        self.addToGroup(self.pi)
        self.addToGroup(self.endi)
        self.addToGroup(self.endf)

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange and self.scene():
            command = CommandMoveItem(self, value)
            self.scene().undoStack.push(command)
        return QGraphicsItemGroup.itemChange(self, change, value)

    def pickle(self):
        pth = []
        for k in range(self.path.elementCount()):
            # e should be either a moveTo or a lineTo
            e = self.path.elementAt(k)
            if e.isMoveTo():
                pth.append(["m", e.x + self.x(), e.y + self.y()])
            else:
                if e.isLineTo():
                    pth.append(["l", e.x + self.x(), e.y + self.y()])
                else:
                    log.error("Problem pickling penarrowitem path {}".format(self.path))
        return ["PenArrow", pth]

    def paint(self, painter, option, widget):
        if not self.scene().itemWithinBounds(self):
            # paint a bounding rectangle out-of-bounds warning
            painter.setPen(QPen(QColor(255, 165, 0), 8))
            painter.setBrush(QBrush(QColor(255, 165, 0, 128)))
            painter.drawRoundedRect(option.rect, 10, 10)
        # paint the normal item with the default 'paint' method
        super(PenArrowItem, self).paint(painter, option, widget)



from plom.db.tables import *

from datetime import datetime

from plom.rules import censorStudentNumber as censorID
from plom.rules import censorStudentName as censorName

import logging

log = logging.getLogger("DB")

from peewee import fn


def getUnknownPageNames(self):
    rval = []
    for uref in UnknownPage.select():
        rval.append(uref.image.file_name)
    return rval


def getDiscardNames(self):
    rval = []
    for dref in DiscardedPage.select():
        rval.append([dref.image.file_name, dref.reason])
    return rval


def getCollidingPageNames(self):
    rval = {}
    for cref in CollidingPage.select():
        rval[cref.image.file_name] = [
            cref.tpage.test.test_number,
            cref.tpage.page_number,
            cref.tpage.version,
        ]
    return rval


def getTPageImage(self, test_number, page_number, version):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    pref = TPage.get_or_none(
        TPage.test == tref,
        TPage.page_number == page_number,
        TPage.version == version,
    )
    if pref is None:
        return [False]
    else:
        return [True, pref.image.file_name]


def getHWPageImage(self, test_number, question, order):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    gref = QGroup.get(test=tref, question=question).group
    pref = HWPage.get_or_none(
        HWPage.test == tref, HWPage.group == gref, HWPage.order == order
    )
    if pref is None:
        return [False]
    else:
        return [True, pref.image.file_name]


def getEXPageImage(self, test_number, question, order):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    gref = QGroup.get(test=tref, question=question).group
    pref = EXPage.get_or_none(
        EXPage.test == tref, EXPage.group == gref, EXPage.order == order
    )
    if pref is None:
        return [False]
    else:
        return [True, pref.image.file_name]


def getLPageImage(self, test_number, order):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    pref = LPage.get_or_none(LPage.test == tref, LPage.order == order)
    if pref is None:
        return [False]
    else:
        return [True, pref.image.file_name]


def getAllTestImages(self, test_number):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    rval = [True]

    # give the pages as IDPages, DNMPages and then for each question

    # grab the id group - only has tpages
    gref = tref.idgroups[0].group
    # give tpages if scanned.
    for p in gref.tpages.order_by(TPage.page_number):
        if p.scanned:
            rval.append(p.image.file_name)

    # grab the dnm group - only has tpages
    gref = tref.dnmgroups[0].group
    # give tpages if scanned.
    for p in gref.tpages.order_by(TPage.page_number):
        if p.scanned:
            rval.append(p.image.file_name)

    # for each question give TPages, HWPages and EXPages
    for qref in tref.qgroups.order_by(QGroup.question):
        gref = qref.group
        for p in gref.tpages.order_by(TPage.page_number):
            if p.scanned:
                rval.append(p.image.file_name)
        for p in gref.hwpages.order_by(HWPage.order):
            rval.append(p.image.file_name)
        for p in gref.expages.order_by(EXPage.order):
            rval.append(p.image.file_name)

    # finally give any loosepages
    for p in tref.lpages.order_by(LPage.order):
        rval.append(p.image.file_name)

    return rval


def getQuestionImages(self, test_number, question):
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]
    qref = QGroup.get_or_none(QGroup.test == tref, QGroup.question == question)
    if qref is None:
        return [False]
    rval = [True]
    # append tpages, hwpages and expages
    for p in qref.group.tpages.order_by(TPage.page_number):
        if p.scanned:
            rval.append(p.image.file_name)
    for p in qref.group.hwpages.order_by(HWPage.order):
        rval.append(p.image.file_name)
    for p in qref.group.expages.order_by(EXPage.order):
        rval.append(p.image.file_name)
    return rval


def getUnknownImage(self, file_name):
    # this really just confirms that the file_name belongs to an unknmown
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:
        return [False]
    uref = iref.upages[0]
    if uref is None:
        return [False]
    else:
        return [True, uref.image.file_name]


def testOwnersLoggedIn(self, tref):
    """Returns list of logged in users who own tasks in given test.

    Note - 'manager' and 'HAL' are not included in this list - else manager could block manager.
    """
    # make list of users who own tasks in the test (might have dupes and 'None')
    user_list = [qref.user for qref in tref.qgroups]
    user_list.append(tref.idgroups[0].user)

    logged_in_list = []
    for uref in user_list:
        if uref:  # make sure uref is not none.
            # avoid adding HAL or manager or duplicates
            if uref.name in ["HAL", "manager"] or uref.name in logged_in_list:
                continue
            if uref.token:
                logged_in_list.append(uref.name)
    return logged_in_list


def moveUnknownToExtraPage(self, file_name, test_number, question):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    uref = iref.upages[0]
    if uref is None:  # should not happen
        return [False, "Cannot find unknown page for that image."]

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False, "Cannot find that test"]
    # check if all owners of tasks in that test are logged out.
    owners = self.testOwnersLoggedIn(tref)
    if owners:
        return [False, "owners", owners]

    # find the qgroup to which the new page should belong
    qref = QGroup.get_or_none(test=tref, question=question)
    if qref is None:  # should not happen
        return [False, "Cannot find that question"]
    version = qref.version  # we'll need the version
    gref = qref.group  # and the parent group
    # find the last expage in that group - if there are expages
    if gref.expages.count() == 0:
        order = 1
    else:
        pref = (
            EXPage.select()
            .where(EXPage.group == gref)
            .order_by(EXPage.order.desc())
            .get()
        )
        order = pref.order + 1

    # now create the expage, delete upage

    with plomdb.atomic():
        xref = EXPage.create(
            test=tref, group=qref.group, version=version, order=order, image=iref
        )
        uref.delete_instance()
        gref.recent_upload = True
        gref.save()
        log.info(
            "Moving unknown page {} to extra page {} of question {} of test {}".format(
                file_name, order, question, test_number
            )
        )
    self.updateTestAfterUpload(tref)
    return [True]


def moveUnknownToHWPage(self, file_name, test_number, question):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    uref = iref.upages[0]
    if uref is None:  # should not happen
        return [False, "Cannot find unknown page for that image."]

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False, "Cannot find that test"]
    # check if all owners of tasks in that test are logged out.
    owners = self.testOwnersLoggedIn(tref)
    if owners:
        return [False, "owners", owners]

    # find the qgroup to which the new page should belong
    qref = QGroup.get_or_none(test=tref, question=question)
    if qref is None:  # should not happen
        return [False, "Cannot find that question"]
    version = qref.version  # we'll need the version
    gref = qref.group  # and the parent group
    # find the last expage in that group - if there are expages
    if gref.hwpages.count() == 0:
        order = 1
    else:
        pref = (
            HWPage.select()
            .where(HWPage.group == gref)
            .order_by(HWPage.order.desc())
            .get()  # there will be at least one
        )
        order = pref.order + 1

    # now create the hwpage, delete upage
    self.createNewHWPage(tref, qref, order, iref)
    uref.delete_instance()
    log.info(
        "Moving unknown page {} to hw page {} of question {} of test {}".format(
            file_name, order, question, test_number
        )
    )
    self.updateTestAfterUpload(tref)
    return [True]


def moveUnknownToTPage(self, file_name, test_number, page_number):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    uref = iref.upages[0]
    if uref is None:  # should not happen
        return [False, "Cannot find unknown page for that image."]

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:  # should not happen
        return [False, "Cannot find that test"]
    # check if all owners of tasks in that test are logged out.
    owners = self.testOwnersLoggedIn(tref)
    if owners:
        return [False, "owners", owners]

    pref = TPage.get_or_none(TPage.test == tref, TPage.page_number == page_number)
    if pref is None:  # should not happen
        return [False, "Cannot find that page."]

    if pref.scanned:
        return [
            False,
            "Page {} of test {} is already scanned.".format(page_number, test_number),
        ]

    self.attachImageToTPage(tref, pref, iref)
    uref.delete_instance()
    log.info(
        "Moving unknown page {} to page {} of test {}".format(
            file_name, page_number, test_number
        )
    )
    self.updateTestAfterUpload(tref)

    return [True]


def checkTPage(self, test_number, page_number):
    """Check whether or not the test/page has been scanned.
    If so then return [collision message, version, image filename]
    Else return [unscanned message, version]
    """
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False, "no such test"]

    pref = TPage.get_or_none(TPage.test == tref, TPage.page_number == page_number)
    if pref is None:
        return [False, "no such page"]
    if pref.scanned:  # we have a collision
        return [True, "collision", pref.version, pref.image.file_name]
    else:  # no collision since the page hasn't been scanned yet
        return [True, "unscanned", pref.version]


def removeUnknownImage(self, file_name):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    uref = iref.upages[0]
    if uref is None:  # should not happen
        return [False, "Cannot find unknown page for that image."]

    with plomdb.atomic():
        DiscardedPage.create(image=iref, reason="User discarded unknown page")
        uref.delete_instance()
    log.info("Discarding unknown image {}".format(file_name))
    return [True]


def getDiscardImage(self, file_name):
    # this really just confirms that the file_name belongs to an discard
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:
        return [False]
    dref = iref.discards[0]
    if dref is None:
        return [False]
    else:
        return [True, dref.image.file_name]


def moveDiscardToUnknown(self, file_name):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    dref = iref.discards[0]
    if dref is None:  # should not happen
        return [False, "Cannot find discard page for that image."]

    with plomdb.atomic():
        UnknownPage.create(image=iref, order=1)  # we have lost order information.
        dref.delete_instance()
    log.info("Moving discarded image {} to unknown image".format(file_name))
    return [True]


def moveUnknownToCollision(self, file_name, test_number, page_number):
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:  # should not happen
        return [False, "Cannot find image"]
    uref = iref.upages[0]
    if uref is None:  # should not happen
        return [False, "Cannot find unknown page for that image."]

    # find the test and the tpage
    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False]

    pref = TPage.get_or_none(TPage.test == tref, TPage.page_number == page_number)
    if pref is None:
        return [False, "Cannot find that page"]
    with plomdb.atomic():
        CollidingPage.create(image=iref, tpage=pref)
        uref.delete_instance()
    log.info(
        "Moving unknown {} to collision of tp {}.{}".format(
            file_name, test_number, page_number
        )
    )
    return [True]


def getCollidingImage(self, file_name):
    # this really just confirms that the file_name belongs to an collidingpage
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:
        return [False]
    cref = iref.collisions[0]
    if cref is None:
        return [False]
    else:
        return [True, cref.image.file_name]


def removeCollidingImage(self, file_name):
    # this really just confirms that the file_name belongs to an collidingpage
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:
        return [False]
    cref = iref.collisions[0]
    if cref is None:
        return [False]

    pref = cref.tpage

    with plomdb.atomic():
        DiscardedPage.create(
            image=iref,
            reason="Removed collision with {}.{}".format(
                pref.test.test_number, pref.page_number
            ),
        )
        cref.delete_instance()
    log.info(
        "Removing collision {} with {}.{}".format(
            file_name, pref.test.test_number, pref.page_number
        )
    )
    return [True]


def moveCollidingToTPage(self, file_name, test_number, page_number, version):
    # this really just confirms that the file_name belongs to an collidingpage
    iref = Image.get_or_none(file_name=file_name)
    if iref is None:
        return [False, "Cannot find image with name {}".format(file_name)]
    cref = iref.collisions[0]
    if cref is None:
        return [False, "Cannot find collision with name {}".format(file_name)]

    tref = Test.get_or_none(Test.test_number == test_number)
    if tref is None:
        return [False, "Cannot find test number {}".format(test_number)]

    pref = TPage.get_or_none(
        TPage.test == tref, TPage.page_number == page_number, TPage.version == version
    )
    if pref is None:
        return [
            False,
            "Cannot find page {} of test {}".format(page_number, test_number),
        ]
    oref = pref.image  # the original page image for this tpage.
    # get the group of that tpage - so we can trigger an update.
    gref = pref.group

    # check if all owners of tasks in that test are logged out.
    owners = self.testOwnersLoggedIn(tref)
    if owners:
        return [False, "owners", owners]

    # now create a discardpage with oref, and put iref into the tpage, delete the collision.
    with plomdb.atomic():
        DiscardedPage.create(
            image=oref,
            reason="Replaced original image {} of {}.{} with new {}".format(
                file_name, pref.test.test_number, pref.page_number, oref.file_name
            ),
        )
        pref.image = iref
        pref.save()
        gref.recent_upload = True
        gref.save()
        cref.delete_instance()
    log.info(
        "Collision {} replacing tpv {}.{}.{}".format(
            file_name, test_number, page_number, version
        )
    )
    # trigger an update since underlying image changed.
    self.updateTestAfterUpload(tref)
    return [True]



# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2019-2020 Colin B. Macdonald
# Copyright (C) 2020 Dryden Wiebe

from pytest import raises

from .utils import rand_hex, rand_integer_code
from .utils import salted_int_hash_from_str as my_hash
from .utils import salted_hex_hash_from_str as hex_hash


def test_hash():
    assert my_hash("12345678", salt="salt", digits=12) == "351525727036"
    assert my_hash("12345678", salt="salty", digits=12) == "782385405730"
    assert my_hash("12345679", salt="salt", digits=12) == "909470548567"


def test_hex_code():
    assert rand_hex().isalnum()
    assert len(rand_hex(1)) == 1
    assert len(rand_hex(2)) == 2
    assert len(rand_hex(32)) == 32


def test_int_code():
    assert isinstance(rand_integer_code(), int)
    assert rand_integer_code(10) > 10


def test_hex_hash_defaults():
    assert len(hex_hash("12345678", salt="salt")) == 16


def test_hash_no_default_salt():
    assert raises(ValueError, lambda: hex_hash("12345678"))
    assert raises(ValueError, lambda: my_hash("123456789"))


def test_hex_hash():
    assert hex_hash("12345678", salt="salt", digits=4) == "1fbe"
    assert hex_hash("12345678", salt="salt", digits=8) == "1fbef050"
    assert hex_hash("12345678", salt="salt", digits=9) == "1fbef0509"
